import logging
import pytest
from psycopg import sql
from psyscale.psql import Commands, Operation, SeriesTbls
from psyscale.psql.enum import AssetTbls

# pylint: disable=missing-function-docstring, protected-access
# --- Mocks ---


def dummy_sql_function(*args, **kwargs):
    return sql.SQL("DUMMY SQL")


@pytest.fixture()
def mock_operation_map():
    return {
        Operation.CREATE: {
            SeriesTbls.TICK: dummy_sql_function,
        },
        Operation.SELECT: {
            SeriesTbls._ORIGIN: dummy_sql_function,
        },
    }


# --- Tests ---


def test_initializes_with_default_map():
    cmds = Commands()
    assert isinstance(cmds.operation_map, dict)
    assert Operation.CREATE in cmds.operation_map

    # Assert access works for both str & strEnum
    assert cmds[Operation.CREATE, AssetTbls.SYMBOLS.value] == cmds[Operation.CREATE, AssetTbls.SYMBOLS]


def test_merge_operations_overrides_existing_strenum(mock_operation_map, caplog):
    caplog.set_level(logging.WARNING)

    cmds = Commands()
    # Copy existing func to compare
    create_tick_original_func = cmds[Operation.CREATE, SeriesTbls.TICK]
    tick_buffer_original_func = cmds[Operation.CREATE, SeriesTbls.TICK_BUFFER]

    # Merge mock which overrides CREATE/TICK
    cmds.merge_operations(mock_operation_map)

    # The original remains unless it was overridden
    assert cmds[Operation.CREATE, SeriesTbls.TICK_BUFFER] == tick_buffer_original_func
    assert cmds[Operation.CREATE, SeriesTbls.TICK] == dummy_sql_function
    assert cmds[Operation.CREATE, SeriesTbls.TICK] != create_tick_original_func

    # Assert Warning Message was Thrown
    assert any(
        f"Overriding psyscale default Operation.CREATE for tables: {set([SeriesTbls.TICK])}" in record.message
        for record in caplog.records
    ), "Expected a warning about overriding CREATE operations"
    assert any(
        f"Overriding psyscale default Operation.SELECT for tables: {set([SeriesTbls._ORIGIN])}" in record.message
        for record in caplog.records
    ), "Expected a warning about overriding SELECT operations"


def test_merge_operations_overrides_existing_str(mock_operation_map, caplog):
    caplog.set_level(logging.WARNING)
    cmds = Commands()
    # Copy existing func to compare
    create_tick_original_func = cmds[Operation.CREATE, "tick"]
    tick_buffer_original_func = cmds[Operation.CREATE, "tick_buffer"]

    # Merge mock which overrides CREATE/TICK
    cmds.merge_operations(mock_operation_map)

    # The original remains unless it was overridden
    assert cmds[Operation.CREATE, "tick_buffer"] == tick_buffer_original_func
    assert cmds[Operation.CREATE, "tick"] == dummy_sql_function
    assert cmds[Operation.CREATE, "tick"] != create_tick_original_func

    # Assert Warning Message was Thrown
    assert any(
        f"Overriding psyscale default Operation.CREATE for tables: {set([SeriesTbls.TICK])}" in record.message
        for record in caplog.records
    ), "Expected a warning about overriding CREATE operations"
    assert any(
        f"Overriding psyscale default Operation.SELECT for tables: {set([SeriesTbls._ORIGIN])}" in record.message
        for record in caplog.records
    ), "Expected a warning about overriding SELECT operations"


def test_custom_map_on_init(mock_operation_map):
    cmds = Commands(operation_map=mock_operation_map)
    assert cmds[Operation.CREATE, SeriesTbls.TICK] == dummy_sql_function


def test_all_return_sql_func():
    cmds = Commands()
    for op, func_pairs in cmds.operation_map.items():
        for tbl in func_pairs.keys():
            func = cmds[op, tbl]
            assert callable(func)


def test_getitem_raises_on_missing_entry():
    cmds = Commands()
    with pytest.raises(ValueError) as excinfo:
        _ = cmds[Operation.CREATE, "value_not_in_op_map"]
    assert "Operation" in str(excinfo.value)



"Metadata Partial Class Functions"

import logging
from typing import Any

from pandas import Timedelta, Timestamp
from psyscale.core import TupleCursor
from psyscale.timeseries_partial import TimeseriesPartialAbstract

from .psql import (
    METADATA_ARGS,
    MetadataArgs,
    MetadataInfo,
    Operation as Op,
    AssetTbls,
    SeriesTbls,
    GenericTbls,
    Schema,
    AssetTable,
)

log = logging.getLogger("psyscale_log")

# pylint: disable='protected-access','abstract-method'


class MetadataPartial(TimeseriesPartialAbstract):
    """
    Series Metadata related sub-routines. Relevant Table initialization handled by
    SymbolsPartial._ensure_securities_schema_format() to ensure symbols table is
    created first.
    """

    def inferred_metadata(self, symbol: int | str, timeframe: Timedelta, rth: bool = False) -> MetadataInfo | None:
        """
        Return the Metadata for a symbol at a given timeframe. This function will use the stored
        metadata to infer what series data can be derived from the stored series data in the event
        the desired timeframe / rth state is not directly stored.
        """
        with self._cursor(dict_cursor=True) as cursor:
            # region ---- Fetch asset_class and stored state
            if isinstance(symbol, str):
                _filters = [("symbol", "ILIKE", symbol)]
            else:
                _filters = [("pkey", "=", symbol)]
            cursor.execute(
                self[Op.SELECT, GenericTbls.TABLE](
                    Schema.SECURITY,
                    AssetTbls.SYMBOLS,
                    ["pkey", "asset_class", "source"],
                    _filters,
                )
            )
            rsp = cursor.fetchall()
            if len(rsp) == 0:
                log.warning("Unknown key : %s", symbol)
                return None
            if len(rsp) > 1:
                log.info("ill defined key, using first result : %s", rsp[0])
            rsp = rsp[0]

            pkey = rsp["pkey"]
            asset_class = rsp["asset_class"]
            assert asset_class

            cursor.execute(self[Op.SELECT, AssetTbls._METADATA](("pkey", "=", pkey)))
            stored_mdata = [MetadataInfo(**row) for row in cursor.fetchall()]

            if len(stored_mdata) == 0:
                log.warning("Key '%s' is known, but not stored.", symbol)
                return None
        # endregion

        _config = self._table_config[Schema(stored_mdata[0].schema_name)]
        desired_table = AssetTable(asset_class, timeframe, False, False, rth)

        # This is the table that will either be selected, or aggregated, from
        # to produce the desired series
        src_table = _config.get_aggregation_source(desired_table, rtn_self=True)
        # Return the metadata associated with the src_table
        return next(mdata for mdata in stored_mdata if mdata.table == src_table)

    def stored_metadata(
        self,
        symbol: int | str,
        filters: dict[MetadataArgs | str, Any] = {},
        *,
        _all: bool = False,
    ) -> list[MetadataInfo]:
        """
        Return Metadata about the series data stored for a given symbol, by ticker or primary key.
        Be default, only returns information about what has been stored & aggregated. Metadata for
        tables that are set to be stored, but are not yet are only returned when _all=True.

        --PARAMS--
        - pkey : integer
            - Primary Key of the desired Symbol. Value is returned as a value in the search_symbols
            return object.
        - filters : Dict[MetadataArgs, Any]
            - Filtering Arguments for the returned MetaData. Extra Keys that don't map to columns
            of the table are ignored. Optional Arguments are as follows.
            - table_name : str
            - schema_name : str
            - is_raw_data : boolean
            - Timeframe : int (Number of seconds elapsed in the period)
            - trading_hours_type : Literal['ext', 'rth', 'eth', 'none']
        - all : boolean
            - When True, Will ignore the given filters and return metadata for what is stored *and*
            what should be stored
            - When False, Will only return metadata for series tables that have data for this
            symbol stored in them
        """

        pkey = self._get_pkey(symbol)
        if pkey is None:
            log.info("Cannot get Metadata, Unknown Symbol: %s", symbol)
            return []

        if _all:
            return _fetch_all_metadata(self, pkey)
        else:
            return _fetch_stored_metadata(self, pkey, filters)

    def manually_refresh_aggregate_metadata(self):
        """
        CLI Script to Manually Refresh Continuous Aggregates as needed.
        Input Options will be presented to completely or partially automate this process
        """
        print(
            """
            Attempting to refresh *all* Continuous Aggregates over *All time*
            This can be an extremely slow process.
              
            The following are options on how to proceed :
              - 'all' - Do everything
              - 'abort' - Do Nothing
              - 'none' - Only refresh the security._metadata view
              - 'schema' - Ask to refresh per schema
              - 'asset' - Ask to refresh per asset_class
              - 'table' - ask to refresh per individual table

            When choosing 'asset' or 'table' the higher level filters will also be available.
        """
        )

        for _ in range(3):
            method = input("'all' / 'none' / 'schema' / 'asset' / 'table' / 'abort' : ")
            if method == "abort":
                return
            if method.lower() in {"all", "none", "schema", "asset", "table"}:
                break
            print("Unknown input")

        if method.lower() not in {"all", "none", "schema", "asset", "table"}:
            print("Learn to type.")
            return

        with self._cursor(auto_commit=True) as cursor:
            if method != "none":
                try:
                    self._manual_refresh_loop(cursor, method)
                except AssertionError:
                    pass

            log.info("---- ---- Refreshing 'Security._Metadata' Materialized View ---- ----")
            cursor.execute(self[Op.REFRESH, AssetTbls._METADATA]())

    def _manual_refresh_loop(self, cursor: TupleCursor, method: str):
        "Inner function that can return but still allow the cursor to refresh the MetaData Table."
        for schema, config in self._table_config.items():
            if method in {"schema", "table", "asset"}:
                rsp = input(f"Refresh schema {schema}? : y/abort/[N] : ")
                if rsp.lower() == "abort":
                    assert False
                if rsp.lower() != "y":
                    continue

            log.info("---- ---- Refreshing Schema : %s ---- ---- ", schema)

            all_aggregates = []

            for asset_class in config.asset_classes:
                if method in {"table", "asset"}:
                    rsp = input(f"Refresh asset_class {asset_class}? : y/abort/[N] : ")
                    if rsp.lower() == "abort":
                        assert False
                    if rsp.lower() != "y":
                        continue

                aggs = config.all_tables(asset_class, include_raw=False)
                aggs.sort(key=lambda x: x.period)
                all_aggregates.extend(aggs)

            for table in all_aggregates:
                if method == "table":
                    rsp = input(f"Refresh table {table}? : y/abort/[N] : ")
                    if rsp.lower() == "abort":
                        assert False
                    if rsp.lower() != "y":
                        continue

                log.info("Refreshing Continuous Aggregate : %s ", table)
                cursor.execute(self[Op.REFRESH, SeriesTbls.CONTINUOUS_AGG](schema, table))


def _fetch_stored_metadata(db: MetadataPartial, pkey: int, filters: dict[MetadataArgs | str, Any] = {}):
    _filters = [("pkey", "=", pkey)]  # Ensure a Pkey filter is Passed
    _filters.extend([(k, "=", v) for k, v in filters.items() if k in (METADATA_ARGS - {"pkey"})])

    rsp, _ = db.execute(db[Op.SELECT, AssetTbls._METADATA](_filters), dict_cursor=True)
    return [MetadataInfo(**row) for row in rsp]


def _fetch_all_metadata(db: MetadataPartial, pkey: int) -> list[MetadataInfo]:
    """
    Fetch MetadataInfo for what is stored *and* what should be stored fpr a given symbol pkey.

    i.e. When A Symbol was just flagged to be stored but has nothing stored yet
    symbol_series_metadata will not return any Metadata since it only checks what *is* stored,
    not what should be. This function will return what is stored + what should be stored.

    If no information is stored for a given table, but should be, the start_date & end_date will
    be set to "1800-01-01" So when a request is made it should automatically fetch all recorded
    data for the symbol.
    """

    _rtn_args = ["asset_class", "store_tick", "store_minute", "store_aggregate"]
    _filter = ("pkey", "=", pkey)

    rsp, _ = db.execute(
        db[Op.SELECT, GenericTbls.TABLE](Schema.SECURITY, AssetTbls.SYMBOLS, _rtn_args, _filter),
        dict_cursor=True,
    )
    if len(rsp) == 0:
        raise ValueError(f"Cannot determine Symbol updates needed. {pkey = } is unknown.")
    rsp = rsp[0]
    asset_class = rsp["asset_class"]

    # A Symbol can only be stored in one schema at a time
    if rsp["store_tick"]:
        schema = Schema.TICK_DATA
    elif rsp["store_minute"]:
        schema = Schema.MINUTE_DATA
    elif rsp["store_aggregate"]:
        schema = Schema.AGGREGATE_DATA
    else:
        log.warning("Requested metadata for Symbol w/ pkey %s, but it is not set to be stored.")
        return []

    try:
        metadata = _fetch_stored_metadata(db, pkey)
        req_tables = db._table_config[schema].raw_tables(asset_class)
        metadata.extend(_missing_metadata(pkey, metadata, req_tables, schema))
    except KeyError as e:
        raise KeyError(  # Reraise a more informative error.
            "Ensure configure_timeseries_schema has been run prior to inserting symbol data."
        ) from e

    return metadata


def _missing_metadata(
    pkey: int,
    stored_metadata: list[MetadataInfo],
    req_tables: list[AssetTable],
    schema: Schema,
) -> list[MetadataInfo]:
    "Determines what metadata is missing, if any, given a list of required tables"
    stored_tables = [mdata.table for mdata in stored_metadata]
    # As long as the hash of an AssetTable is a string this will work.
    missing_tables = set(req_tables).difference(stored_tables)
    missing_metadata = [
        MetadataInfo(
            pkey,
            table.table_name,
            schema,
            Timestamp("1800-01-01", tz="UTC"),  # Default values
            Timestamp("1800-01-01", tz="UTC"),
            table,
        )
        for table in missing_tables
    ]
    return missing_metadata



"""
A simple Python library for storing & retrieving financial data
in a TimescaleDB optimized Postgres Database.
"""

import asyncio
import logging
import sys
from typing import Literal


from .psql import AssetTable, TimeseriesConfig, MetadataInfo
from .core import PsyscaleConnectParams
from .symbols_partial import AsyncSymbolsPartial, SymbolsPartial
from .timeseries_partial import TimerseriesPartial, TimeseriesAsyncPartial


class PsyscaleDB(TimerseriesPartial, SymbolsPartial):
    """
    Synchronous client interface for connecting to a PostgreSQL + TimescaleDB Database.
    Timescale DB Docker self-host instructions
    https://docs.timescale.com/self-hosted/latest/install/installation-docker/
    """


class PsyscaleAsync(TimeseriesAsyncPartial, AsyncSymbolsPartial):
    """
    Asynchronous client interface for connecting to a PostgreSQL + TimescaleDB Database.

    Timescale DB Docker self-host instructions
    https://docs.timescale.com/self-hosted/latest/install/installation-docker/
    """


__all__ = (
    "PsyscaleDB",
    "PsyscaleAsync",
    "PsyscaleConnectParams",
    "set_psyscale_log_level",
    "AssetTable",
    "TimeseriesConfig",
    "MetadataInfo",
)


_log = logging.getLogger("psyscale_log")
handler = logging.StreamHandler(None)
formatter = logging.Formatter("[PsyscaleDB] - [.\\%(filename)s Line: %(lineno)d] - %(levelname)s: %(message)s")
handler.setFormatter(formatter)
_log.addHandler(handler)
_log.setLevel("WARNING")


def set_psyscale_log_level(
    level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
):
    "Set the logging Level for all TimescaleDB Logs."
    _log.setLevel(level)


if sys.platform == "win32":
    _log.info("Setting Asyncio Loop Policy to be compatible with asyncio + psycopg3")
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())



"Tests the Asyncronous method versions for Symbol and Data insertion."

import pandas as pd
from pandas.testing import assert_series_equal

import pytest

from psyscale import PsyscaleAsync
from psyscale.dev import TimeseriesConfig
from psyscale.psql.enum import Schema
from psyscale.psql.timeseries import AGGREGATE_ARGS
from psyscale.series_df import Series_DF


@pytest.fixture(scope="module")
def AAPL_MIN_DATA():
    df = pd.read_csv("example_data/AAPL_1min.csv")
    df.rename({"dt": "date", "open": "o", "high": "max"}, inplace=True)
    yield df


@pytest.fixture(scope="module")
def symbols():
    yield pd.DataFrame(
        [
            {
                "symbol": "AAPL",
                "name": "Apple Inc.",
                "exchange": "NASDAQ",
                "asset_class": "equity",
                "sector": "Technology",
            },
            {
                "symbol": "GOOG",
                "name": "Alphabet Inc.",
                "exchange": "NASDAQ",
                "asset_class": "equity",
                "sector": "Technology",
            },
        ]
    )


MINUTE_CONFIG = TimeseriesConfig(
    ["equity"],  # type:ignore
    rth_origins={
        "equity": pd.Timestamp("2000/01/03 08:30", tz="America/New_York"),
    },
    eth_origins={
        "equity": pd.Timestamp("2000/01/03 04:00", tz="America/New_York"),
    },
    prioritize_rth={"equity": True},
    calculated_periods={"default": [pd.Timedelta("5m"), pd.Timedelta("30m"), pd.Timedelta("1h")]},
    stored_periods={"default": [pd.Timedelta("1m")]},
)


def test_00_configure(psyscale_async: PsyscaleAsync):
    psyscale_async.configure_timeseries_schema(minute_tables=MINUTE_CONFIG)


@pytest.mark.asyncio
async def test_01_insert_symbols(psyscale_async: PsyscaleAsync, symbols):
    await psyscale_async.upsert_securities_async(symbols, source="unit_test")

    rsp = await psyscale_async.search_symbols_async({"asset_class": "equity"})
    assert len(rsp) == 2

    rsp = await psyscale_async.search_symbols_async({"symbol": "AAPL"}, strict_symbol_search=True)
    aapl = rsp[0]
    assert aapl["pkey"]
    assert aapl["symbol"] == "AAPL"

    # No Async Version of this function
    psyscale_async.update_symbol(aapl["pkey"], {"store_minute": True})


@pytest.mark.asyncio
async def test_01_metadata_fetch(psyscale_async: PsyscaleAsync):
    rsp = await psyscale_async.search_symbols_async({"symbol": "aapl"})
    assert len(rsp) == 1
    aapl = rsp[0]
    assert aapl["symbol"] == "AAPL"
    assert "pkey" in aapl
    aapl_pkey = aapl["pkey"]

    # No Data inserted yet, MetaData should be empty, No Async Metadata function
    metadata = psyscale_async.stored_metadata(aapl_pkey)
    assert len(metadata) == 0

    # Should show that we need to store data in the Minute table of minute schema
    metadata = psyscale_async.stored_metadata(aapl_pkey, _all=True)
    assert len(metadata) == 1
    metadata = metadata[0]

    assert metadata.schema_name == Schema.MINUTE_DATA
    assert metadata.start_date == metadata.end_date
    assert metadata.start_date == pd.Timestamp("1800-01-01", tz="UTC")
    assert metadata.timeframe == pd.Timedelta("1min")


@pytest.mark.asyncio
async def test_02_aggregate_data_insert(psyscale_async: PsyscaleAsync, caplog, AAPL_MIN_DATA):
    # pkey and metadata fetch already asserted working in first test.
    aapl = (await psyscale_async.search_symbols_async({"store_minute": True}))[0]
    metadata = psyscale_async.stored_metadata(aapl["pkey"], _all=True)[0]

    with pytest.raises(ValueError):
        # Should error since a 'rth' column is needed but not given.
        await psyscale_async.upsert_series_async(aapl["pkey"], metadata, AAPL_MIN_DATA, None)

    with pytest.raises(AttributeError):
        # Should error since a column is missing
        await psyscale_async.upsert_series_async(aapl["pkey"], metadata, AAPL_MIN_DATA.drop(columns="date"), None)

    # These inter trackers should not yet be populated
    assert not hasattr(psyscale_async, "_altered_tables")
    assert not hasattr(psyscale_async, "_altered_tables_mdata")

    # Should Work, renames columns and all (column renaming tests in series_df tests)
    # And Dropps Extra Columns
    with caplog.at_level("DEBUG"):
        await psyscale_async.upsert_series_async(aapl["pkey"], metadata, AAPL_MIN_DATA, "NYSE")

    # Assert there was no cursor error in the upsert function.
    assert all(record.levelname != "ERROR" for record in caplog.records)

    # Now they should be populated
    assert hasattr(psyscale_async, "_altered_tables")
    assert hasattr(psyscale_async, "_altered_tables_mdata")

    with caplog.at_level("INFO"):
        await psyscale_async.refresh_aggregate_metadata_async()

    # Assert tracking variables are cleaned up
    assert not hasattr(psyscale_async, "_altered_tables")
    assert not hasattr(psyscale_async, "_altered_tables_mdata")

    # check what's available
    metadata = psyscale_async.stored_metadata(aapl["pkey"])
    assert len(metadata) == 4  # One for inserted data, 3 more for the aggregates


@pytest.mark.asyncio
async def test_03_check_inserted_data(psyscale_async: PsyscaleAsync):
    aapl = (await psyscale_async.search_symbols_async({"store_minute": True}))[0]

    # check what's available
    metadata = psyscale_async.stored_metadata(aapl["pkey"])
    assert len(metadata) == 4  # One for inserted data, 3 more for the aggregates

    minute_metadata = [m for m in metadata if m.timeframe == pd.Timedelta("1min")][0]

    # check that the full dataset got inserted
    raw_data = Series_DF(pd.read_csv("example_data/AAPL_1min.csv"), "NYSE")
    raw_data.df.drop(columns=set(raw_data.columns).difference(AGGREGATE_ARGS), inplace=True)
    raw_data.df.set_index(keys=pd.RangeIndex(0, 2083), inplace=True)
    raw_data.df["rth"] = raw_data.df["rth"].astype("int64")

    inserted_data = await psyscale_async.get_series_async(
        aapl["pkey"],
        pd.Timedelta("1min"),
        rth=False,
        rtn_args={"open", "high", "low", "close", "volume", "rth"},
        mdata=minute_metadata,
    )
    assert inserted_data is not None

    for col in raw_data.columns:
        if col == "volume":
            assert_series_equal(raw_data.df[col].astype("int64"), inserted_data[col])
        else:
            assert_series_equal(raw_data.df[col], inserted_data[col])

    # check that only the 'rth' data cna be retrieved
    raw_data.df = raw_data.df[raw_data.df["rth"] == 0]
    raw_data.df.set_index(keys=pd.RangeIndex(0, 780), inplace=True)

    inserted_data = await psyscale_async.get_series_async(
        aapl["pkey"],
        pd.Timedelta("1min"),
        rth=True,
        rtn_args={"open", "high", "low", "close", "volume", "rth"},
    )
    assert inserted_data is not None

    for col in raw_data.columns:
        if col == "volume":
            assert_series_equal(raw_data.df[col].astype("int64"), inserted_data[col])
        else:
            assert_series_equal(raw_data.df[col], inserted_data[col])



"""An interface for reading and commiting data to a database"""

from asyncio import create_task
import asyncio
import logging
from inspect import stack
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import (
    AsyncIterator,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    TypeAlias,
    overload,
)
import psycopg as pg
import psycopg.rows as pg_rows
from psycopg import AsyncConnection, OperationalError, sql
from psycopg.pq._enums import ExecStatus
from psycopg_pool import AsyncConnectionPool, PoolTimeout

from .core import (
    PsyscaleConnectParams,
    POOL_GEN_TIMEOUT,
    LOCAL_POOL_GEN_TIMEOUT,
    PsyscaleCore,
)

log = logging.getLogger("psyscale_log")

AsyncDictCursor: TypeAlias = pg.AsyncCursor[pg_rows.DictRow]
AsyncTupleCursor: TypeAlias = pg.AsyncCursor[pg_rows.TupleRow]


class ImproperInitilizationError(Exception):
    "Class not initialized through required create() class method."


class PsyscaleAsyncCore(PsyscaleCore):
    """
    Core Asynchronous client interface for connecting to a PostgreSQL Database.
    Extends the PsyscaleCore Client with an Async Connection Pool to allow for
    a Hybrid Requests.

    Hybrid approach is taken since some functions, when asynchronously awaited,
    will result in more overhead than what is gained by allowing other processes
    to run in the await time.
    """

    def __init__(
        self,
        conn_params: PsyscaleConnectParams | str | None = None,
        *,
        down_on_del: bool = False,
        docker_compose_fpath: str | None = None,
    ):
        super().__init__(
            conn_params,
            down_on_del=down_on_del,
            docker_compose_fpath=docker_compose_fpath,
        )
        self._async_pool: AsyncConnectionPool
        self._pool_task = create_task(self._open())
        asyncio.gather(create_task(self._async_health_check())).add_done_callback(
            lambda good: log.debug(
                "Async Pool Health_check: %s",
                "good" if good else "bad",
            )
        )

    @property
    async def _async_conn(self) -> AsyncConnection:
        if not self._pool_task.done():
            await asyncio.gather(self._pool_task)
        return await self._async_pool.getconn()

    async def close(self):
        "Close the active connection pool prior to __del__."
        await self._async_pool.close()

    async def _open(self):
        "Open the async connection pool. Task automatically created on init."
        _timeout = LOCAL_POOL_GEN_TIMEOUT if self.db_cfg.is_local else POOL_GEN_TIMEOUT

        try:
            self._async_pool = AsyncConnectionPool(self.db_cfg.url, open=False, timeout=_timeout)
            await self._async_pool.open(timeout=_timeout)
        except PoolTimeout as e:
            raise e  #
        except OperationalError as e:
            raise e  # Give some more informative info here?

    # region ----------- Async Connection & Cursor Methods -----------

    # region -- Cursor & Execute Overloading ---

    @overload
    def _acursor(
        self,
        dict_cursor: Literal[True],
        *,
        binary: bool = False,
        pipeline: bool = False,
        raise_err: bool = False,
        auto_commit: bool = False,
    ) -> AbstractAsyncContextManager[AsyncDictCursor]: ...
    @overload
    def _acursor(
        self,
        dict_cursor: Literal[False] = False,
        *,
        binary: bool = False,
        pipeline: bool = False,
        raise_err: bool = False,
        auto_commit: bool = False,
    ) -> AbstractAsyncContextManager[AsyncTupleCursor]: ...
    @overload
    def _acursor(
        self,
        dict_cursor: bool = False,
        *,
        binary: bool = False,
        pipeline: bool = False,
        raise_err: bool = False,
        auto_commit: bool = False,
    ) -> AbstractAsyncContextManager[AsyncTupleCursor]: ...

    @overload
    async def execute_async(
        self,
        cmd: sql.Composed,
        exec_args: Optional[Mapping[str, int | float | str | None]] = None,
        dict_cursor: Literal[False] = False,
    ) -> Tuple[List[Tuple], str | None]: ...
    @overload
    async def execute_async(
        self,
        cmd: sql.Composed,
        exec_args: Optional[Mapping[str, int | float | str | None]] = None,
        dict_cursor: Literal[True] = True,
    ) -> Tuple[List[Dict], str | None]: ...

    # endregion

    @asynccontextmanager
    async def _acursor(
        self,
        dict_cursor: bool = False,
        *,
        binary: bool = False,
        pipeline: bool = False,
        raise_err: bool = False,
        auto_commit: bool = False,
    ) -> AsyncIterator[AsyncTupleCursor | AsyncDictCursor]:
        """
        Returns a cursor to execute commands in a database.

        Default return product is a list of tuples. Returns can be made into lists of dictionaries
        by settign dict_cursor=True. This is less performant for large datasets though.

        Auto_Commit Allows for commands to be done outside of a transaction block which may be
        required for some commands.

        Pipeline is a feature of a cursor, that when set to True, avoids waiting for responses
        before executing new commands. In theory that should increase performance. In practice
        it seemed to only reduce performance.

        raise_err = True Raises any database errors after rolling back the database.
        raise_err = False Rollsback changes, logs an error, then silences the error.
        """
        cursor_factory = pg_rows.dict_row if dict_cursor else pg_rows.tuple_row
        conn: pg.AsyncConnection = await self._async_conn

        if auto_commit:
            await conn.set_autocommit(True)

        try:
            if pipeline:
                async with (
                    conn.pipeline(),
                    conn.cursor(row_factory=cursor_factory, binary=binary) as cursor,
                ):
                    yield cursor  # type:ignore : Silence the Dict/Tuple overloading Error
            else:
                async with (
                    conn,
                    conn.cursor(row_factory=cursor_factory, binary=binary) as cursor,
                ):
                    yield cursor  # type:ignore : Silence the Dict/Tuple overloading Error
        except pg.DatabaseError as e:
            await conn.rollback()  # Reset Database, InFailedSqlTransaction Err thrown if not reset
            log.error("Caught Database Error: \n '%s' \n...Rolling back changes.", e)

            if raise_err:
                raise e  # log the message and continue to let the error bubble
        finally:
            if auto_commit:
                await conn.set_autocommit(False)
            else:
                await conn.commit()

            # Pool is guaranteed to exist at this point
            await self._async_pool.putconn(conn)

    async def execute_async(
        self,
        cmd: sql.Composed | sql.SQL,
        exec_args: Optional[Mapping[str, int | float | str | None]] = None,
        dict_cursor: bool = False,
    ) -> Tuple[List[Dict] | List[Tuple], str | None]:
        """
        Execution Method to manually invoke a command within the database.

        -- PARAMS --
        - cmd : sql.Composed
            - Composed SQL function using psycopg.sql sub-module
        - exec_args : Optional Mapping
            - When supplied, exec_args will be passed to the cursor and used to populate
            any placeholder args within the formatted command.
        - dict_cursor : boolean
            - When true will return any results as a list of dicts where each item in
            the list is a row of the returned table.
            - When false a list of Tuples per row is returned.

        -- RETURNS --
        Tuple[ List[] , str ] ==> Tuple of query results (if any) and the cursor status
        message returned as a string.
        """
        async with self._acursor(dict_cursor) as cursor:
            try:
                log.debug("Executing PSQL Command: %s", cmd.as_string(cursor))
                await cursor.execute(cmd, exec_args)

            except pg.DatabaseError as e:
                log.error(
                    "Cursor Execute Exception (%s) occured in '%s' \n  Exception Msg: %s",
                    e.__class__.__name__,
                    stack()[1].function,
                    e,
                )
                return [], cursor.statusmessage

            response = []
            pgr = cursor.pgresult
            # Really sad I have to dig to check if there is data available.
            if pgr is not None and pgr.status == ExecStatus.TUPLES_OK:
                response = await cursor.fetchall()

            return response, cursor.statusmessage

    async def _async_health_check(self) -> bool:
        "Simple Ping to the Database to ensure it is alive"
        async with self._acursor() as cursor:
            await cursor.execute("SELECT 1")
            rsp = await cursor.fetchall()
            return rsp[0][0] == 1
        return False

    # endregion



import pytest
from pandas import Timedelta, Timestamp
from psyscale.dev import AssetTable

# region ---- ---- Asset Table ---- ----


def test_asset_table_table_name_variants():
    at = AssetTable("stock", Timedelta("5min"), raw=True, ext=True, rth=None)
    assert at.table_name == "stock_300_raw_ext"

    at_rth = AssetTable("crypto", Timedelta("5min"), raw=True, ext=True, rth=True)
    assert at_rth.table_name == "crypto_300_raw_rth"

    at_eth = AssetTable("stock", Timedelta("5min"), raw=True, ext=True, rth=False)
    assert at_eth.table_name == "stock_300_raw_eth"

    at_nonext = AssetTable("etf", Timedelta("5min"), raw=True, ext=False, rth=None)
    assert at_nonext.table_name == "etf_300_raw"

    at_agg = AssetTable("stock", Timedelta("5min"), raw=False, ext=False, rth=None)
    assert at_agg.table_name == "stock_300"


def test_asset_table_equality_repr_hash():
    at1 = AssetTable("crypto", Timedelta("1h"), raw=False, ext=True, rth=None)
    at2 = AssetTable("crypto", Timedelta("1h"), raw=False, ext=True, rth=None)
    assert at1 == at2
    assert hash(at1) == hash(at2)

    at = AssetTable("equity", Timedelta("1d"), raw=False, ext=False, rth=None)
    assert repr(at) == "equity_86400"


def test_asset_table_properties():
    at = AssetTable("etf", Timedelta("3d"), raw=False, ext=True, rth=None)
    assert isinstance(at.origin_ts, Timestamp)
    assert isinstance(at.origin_ltf, str)
    assert isinstance(at.origin_htf, str)
    assert at.psql_interval == "259200 seconds"

    # Has RTH Cases
    assert at.has_rth is True
    assert AssetTable("etf", Timedelta("3d"), raw=True, ext=True, rth=False).has_rth
    assert not AssetTable("etf", Timedelta("3d"), raw=False, ext=True, rth=True).has_rth
    assert not AssetTable("etf", Timedelta("3d"), raw=False, ext=False, rth=None).has_rth


def test_asset_table_from_table_name():
    table_name = "futures_3600_raw_ext"
    at = AssetTable.from_table_name(table_name)
    assert at.asset_class == "futures"
    assert at.period == Timedelta(seconds=3600)
    assert at.raw is True
    assert at.ext is True
    assert at.rth is None


def test_asset_table_origin_fallback():
    at_ltf = AssetTable("test", Timedelta("3d"), raw=False, ext=False, rth=None)
    assert at_ltf.origin == str(at_ltf.origin_ts)

    at_htf = AssetTable("test", Timedelta("6W"), raw=False, ext=False, rth=None)
    assert at_htf.origin == str(at_htf.origin_ts)


def test_interval_limit():
    with pytest.raises(ValueError):
        _ = AssetTable("test", Timedelta("1856.5s"), raw=False, ext=False, rth=None)


# endregion



import pytest
from psycopg import sql

from psyscale.psql.generic import (
    where,
    limit,
    order,
    arg_list,
    update_args,
    filter_composer,
)


def test_where():
    # Filter Literal
    stmt = where(("id", "=", 5))
    assert stmt.as_string() == ' WHERE "id"=5'
    # Filter Placeholder
    stmt = where(("name", "ILIKE"))
    assert stmt.as_string() == ' WHERE "name"ILIKE%(name)s'
    # Empty list
    assert where([]).as_string() == ""


def test_limit_with_str():
    assert limit(10).as_string() == " LIMIT 10"
    assert limit("param").as_string() == " LIMIT %(param)s"
    assert limit(None).as_string() == ""


def test_order_default():
    stmt = order("created_at")
    assert stmt.as_string() == ' ORDER BY "created_at"'
    stmt = order("created_at", ascending=False)
    assert stmt.as_string() == ' ORDER BY "created_at" DESC'
    stmt = order(None)
    assert stmt.as_string() == ""


def test_arg_list():
    result = arg_list(["a", "b"]).as_string()
    assert result == '"a", "b"'
    result = arg_list(["a", "b"], distinct=True).as_string()
    assert result == 'DISTINCT "a", "b"'
    assert arg_list([]).as_string() == "*"
    assert arg_list([], distinct=True).as_string() == "DISTINCT *"


def test_update_args():
    args = [("a", 1), ("b", True)]  # Literals
    result = update_args(args).as_string()
    assert result == '"a" = 1, "b" = true'

    args = ["a", "b"]  # Placeholders
    result = update_args(args).as_string()
    assert result == '"a" = %(a)s, "b" = %(b)s'

    args = [("a", 1), "b"]  # Literals & Placeholders
    result = update_args(args).as_string()
    assert result == '"a" = 1, "b" = %(b)s'

    with pytest.raises(ValueError):
        update_args([])


def test_filter_composer_literals():
    filters = [("a", "=", 1), ("b", "LIKE", "%abc%")]  # Literals
    result = filter_composer(filters).as_string()
    assert result == """"a"=1 AND "b"LIKE'%abc%'"""

    filters = [("a", ">"), ("b", "ILIKE")]  # Placeholders
    # Following line can't recognize the Comparator Literals, apparently thinks they're strings
    result = filter_composer(filters).as_string()  # type: ignore
    assert result == '"a">%(a)s AND "b"ILIKE%(b)s'

    filters = [sql.SQL("EXISTS(SELECT 1)")]  # Inner SQL
    result = filter_composer(filters).as_string()
    assert result == "(EXISTS(SELECT 1))"

    filters = [("a", "!=", 1), ("b", "=", 2)]
    result = filter_composer(filters, mode="OR").as_string()  # type: ignore
    assert result == '"a"!=1 OR "b"=2'



"""Formatted SQL Commands for all Tables & Views Under the Timeseries Schemas"""

from typing import Literal, Optional, get_args

from pandas import Timedelta, Timestamp
from psycopg import sql

from .orm import AssetTable, MetadataInfo
from .enum import SeriesTbls, AssetTbls, Schema
from .generic import Filter, where, limit, arg_list, order

# pylint: disable=missing-function-docstring, protected-access, line-too-long


# region -------- -------- Timeseries Bucket Origins Commands -------- --------
# A Table of Timestamps that Store Time bucket origins for each given asset class.


def create_origin_table(schema: str) -> sql.Composed:
    "A Table of Timestamps that Store Time bucket origins for each given asset class."
    return sql.SQL(
        """
        CREATE TABLE {schema_name}.{table_name} (
            asset_class TEXT PRIMARY KEY,
            origin_rth TIMESTAMPTZ NOT NULL,
            origin_eth TIMESTAMPTZ NOT NULL,
            origin_htf TIMESTAMPTZ NOT NULL
        );
    """
    ).format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(SeriesTbls._ORIGIN),
    )


def select_origin(
    schema: str,
    asset_class: str = "",
    rth: bool | None = None,
    period: Timedelta = Timedelta(-1),
    *,
    _all: bool = False,
) -> sql.Composed:
    if _all:
        return _select_all_origins(schema)
    else:
        return _select_origin(schema, asset_class, rth, period) + sql.SQL(";")


def _select_all_origins(schema: str) -> sql.Composed:
    return sql.SQL("SELECT (asset_class, origin_rth, origin_eth, origin_htf) FROM {schema_name}.{table_name};").format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(SeriesTbls._ORIGIN),
    )


def _select_origin(
    schema: str,
    asset_class: str = "",
    rth: bool | None = None,
    period: Timedelta = Timedelta(-1),
) -> sql.Composed:
    origin = "origin_htf" if period >= Timedelta("4W") else "origin_rth" if rth else "origin_eth"
    return sql.SQL("SELECT {origin} FROM {schema_name}.{table_name} WHERE asset_class = {asset_class}").format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(SeriesTbls._ORIGIN),
        origin=sql.Identifier(origin),
        asset_class=sql.Literal(asset_class),
    )


def insert_origin(
    schema: str,
    asset_class: str,
    rth_origin: Timestamp,
    eth_origin: Timestamp,
    htf_origin: Timestamp,
) -> sql.Composed:
    return sql.SQL(
        "INSERT INTO {schema_name}.{table_name} (asset_class, origin_rth, origin_eth, origin_htf) "
        "VALUES ({asset_class}, {origin_rth}, {origin_eth}, {origin_htf});"
    ).format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(SeriesTbls._ORIGIN),
        asset_class=sql.Literal(asset_class),
        origin_rth=sql.Literal(str(rth_origin)),
        origin_eth=sql.Literal(str(eth_origin)),
        origin_htf=sql.Literal(str(htf_origin)),
    )


def update_origin(
    schema: str,
    asset_class: str,
    rth_origin: Timestamp,
    eth_origin: Timestamp,
    htf_origin: Timestamp,
) -> sql.Composed:
    return sql.SQL(
        """
        UPDATE {schema_name}.{table_name} SET
            origin_rth = {origin_rth},
            origin_eth = {origin_eth},
            origin_htf = {origin_htf}
        WHERE asset_class = {asset_class};
        """
    ).format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(SeriesTbls._ORIGIN),
        asset_class=sql.Literal(asset_class),
        origin_rth=sql.Literal(str(rth_origin)),
        origin_eth=sql.Literal(str(eth_origin)),
        origin_htf=sql.Literal(str(htf_origin)),
    )


def delete_origin(schema: str, asset_class: str) -> sql.Composed:
    return sql.SQL("DELETE FROM {schema_name}.{table_name} WHERE asset_class = {asset_class};").format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(SeriesTbls._ORIGIN),
        asset_class=sql.Literal(asset_class),
    )


# endregion


# region -------- -------- Tick Timeseries Commands -------- --------

TickArgs = Literal[
    "dt",
    "rth",
    "price",
    "volume",
]
TICK_ARGS = set(v for v in get_args(TickArgs))


def create_tick_table(schema: str, table: AssetTable) -> sql.Composed:
    "Create a Tick table and the initial aggregate needed for other aggregates"
    if table.period != Timedelta(0):
        raise ValueError("A Tick Table must have a Period of Timedelta(0).")
    return sql.SQL(
        """
        CREATE TABLE {schema_name}.{table_name} (
            pkey INTEGER NOT NULL,
            dt TIMESTAMPTZ NOT NULL,"""
        + (" rth SMALLINT not NULL," if table.has_rth else "")
        + """
            price DOUBLE PRECISION NOT NULL,
            volume DOUBLE PRECISION, 
            PRIMARY KEY (pkey, dt),
            CONSTRAINT fk_pkey FOREIGN KEY (pkey) REFERENCES {ref_schema_name}.{ref_table_name} (pkey)
        );
        SELECT create_hypertable({full_table_name}, by_range('dt'));
        """
    ).format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(str(table)),
        full_table_name=sql.Literal(schema + "." + repr(table)),
        # There's no real good way to make the reference table dynamic
        # Leaving these as Identifies so its easier to see these variable names are hard-coded
        ref_schema_name=sql.Identifier(Schema.SECURITY),
        ref_table_name=sql.Identifier(AssetTbls.SYMBOLS),
    )


def create_continuous_tick_aggregate(schema: str, table: AssetTable, ref_table: AssetTable) -> sql.Composed:
    "Create the inital continuous aggregate from a tick table."
    _error_check_continuous_aggrigate(table, ref_table)
    return sql.SQL(
        """
        CREATE MATERIALIZED VIEW {schema_name}.{table_name} WITH (timescaledb.continuous) AS
        SELECT 
            time_bucket({interval}, dt, TIMESTAMPTZ {origin}) as dt,
            pkey,"""
        + ("    first(rth, dt) AS rth," if table.has_rth else "")
        + """
            first(price, dt) AS open,
            last(price, dt) AS close,
            max(price) AS high,
            min(price) AS low,
            sum(volume) AS volume,
            sum(price * volume) / NULLIF(SUM(volume), 0) AS vwap,
            count(*) AS ticks
        FROM {schema_name}.{ref_table_name}"""
        # Where clause handles the case when going from a table with an rth column to no rth column
        + (" WHERE rth = 0 " if ref_table.rth is None and table.ext and table.rth else "")
        + """
        GROUP BY pkey, 1 ORDER BY 1
        WITH NO DATA;
        """
    ).format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(repr(table)),
        ref_table_name=sql.Identifier(repr(ref_table)),
        origin=sql.Literal(table.origin),
        interval=sql.Literal(table.psql_interval),
    )


def _error_check_continuous_aggrigate(table: AssetTable, src_table: AssetTable):
    if table.ext != src_table.ext:
        # One table has extended hours information, the other doesn't
        raise AttributeError(
            "EXT Data mismatched between reference and continuous aggregate tables.\n"
            f"Desired Aggregate Table: {table}, Reference Table: {src_table}"
        )
    if not src_table.ext or src_table.rth is None:
        return  # either ext = None for both, or src_table has all ext information needed
    if table.rth is None:  # ref_table.rth == True or False
        # Trying to get more refined ext information out of src_table than what is possible
        raise AttributeError(
            "Cannot create an Aggregate table with an rth column from a table with an rth column."
            f"Desired Aggregate Table: {table}, Reference Table: {src_table}"
        )
    if src_table.rth != table.rth:
        raise AttributeError(
            "Cannot create an RTH Aggregate from an ETH aggregate and vise versa."
            f"Desired Aggregate Table: {table}, Reference Table: {src_table}"
        )


def refresh_continuous_aggregate(
    schema: str, table: AssetTable, start: Optional[Timestamp], end: Optional[Timestamp]
) -> sql.Composed:
    if table.raw:
        raise AttributeError(f"Cannot Refresh Table {schema}.{table}. It is not a Continuous Aggregate.")
    return sql.SQL("CALL refresh_continuous_aggregate({full_name}, {start}, {end});").format(
        full_name=sql.Literal(schema + "." + repr(table)),
        start=sql.Literal(start),  # Automatically handles type conversion & Null Case
        end=sql.Literal(end),
    )


def create_raw_tick_buffer(table: AssetTable) -> sql.Composed:
    return sql.SQL(
        """
        CREATE TEMP TABLE _tick_buffer (
            dt TIMESTAMPTZ NOT NULL,"""
        + (" rth SMALLINT not NULL," if table.has_rth else "")
        + """
            price DOUBLE PRECISION NOT NULL,
            volume DOUBLE PRECISION DEFAULT NULL
        ) ON COMMIT DROP;
    """
    ).format()


def copy_ticks(args: list[str]) -> sql.Composed:
    return sql.SQL("COPY _tick_buffer ({args}) FROM STDIN;").format(
        args=sql.SQL(",").join([sql.Identifier(arg) for arg in args]),
    )


def insert_copied_ticks(schema: str, table: AssetTable, pkey: int) -> sql.Composed:
    "No Conflict Statement since Inserted Data Ideally should be only new data."
    if table.has_rth:
        return sql.SQL(
            """
            INSERT INTO {schema_name}.{table_name} (pkey, dt, rth, price, volume) 
            SELECT {pkey}, dt, rth, price, volume FROM _tick_buffer;
        """
        ).format(
            schema_name=sql.Identifier(schema),
            table_name=sql.Identifier(str(table)),
            pkey=sql.Literal(pkey),
        )
    else:
        return sql.SQL(
            """
            INSERT INTO {schema_name}.{table_name} (pkey, dt, price, volume) 
            SELECT {pkey}, dt, price, volume FROM _tick_buffer;
        """
        ).format(
            schema_name=sql.Identifier(schema),
            table_name=sql.Identifier(str(table)),
            pkey=sql.Literal(pkey),
        )


def upsert_copied_ticks(schema: str, table: AssetTable, pkey: int) -> sql.Composed:
    "Not Intended to be used as often as INSERT Operation since this requires a CONTINUOUS AGG REFRESH"
    if table.has_rth:
        return sql.SQL(
            """
            INSERT INTO {schema_name}.{table_name} (pkey, dt, rth, price, volume) 
            SELECT {pkey}, dt, rth, price, volume FROM _tick_buffer
            ON CONFLICT (pkey, dt) DO UPDATE
            SET price = EXCLUDED.price, volume = EXCLUDED.volume, rth = EXCLUDED.rth;
        """
        ).format(
            schema_name=sql.Identifier(schema),
            table_name=sql.Identifier(str(table)),
            pkey=sql.Literal(pkey),
        )
    else:
        return sql.SQL(
            """
            INSERT INTO {schema_name}.{table_name} (pkey, dt, price, volume) 
            SELECT {pkey}, dt, price, volume FROM _tick_buffer
            ON CONFLICT (pkey, dt) DO UPDATE
            SET price = EXCLUDED.price, volume = EXCLUDED.volume;
        """
        ).format(
            schema_name=sql.Identifier(schema),
            table_name=sql.Identifier(str(table)),
            pkey=sql.Literal(pkey),
        )


# endregion


# region -------- -------- -------- Aggrigate Timeseries Commands -------- -------- --------

AggregateArgs = Literal["dt", "rth", "open", "high", "low", "close", "volume", "vwap", "ticks"]
AGGREGATE_ARGS = set(v for v in get_args(AggregateArgs))


def create_raw_aggregate_table(schema: str, table: AssetTable) -> sql.Composed:
    "Aggregate Table that is filled with data from a source API, Should be maintained by the user."
    return sql.SQL(
        """
        CREATE TABLE {schema_name}.{table_name} (
            pkey INTEGER NOT NULL,
            dt TIMESTAMPTZ NOT NULL,"""
        + (" rth SMALLINT NOT NULL," if table.has_rth else "")
        + """
            close DOUBLE PRECISION NOT NULL,
            open DOUBLE PRECISION,
            high DOUBLE PRECISION,
            low DOUBLE PRECISION,
            volume DOUBLE PRECISION,
            vwap DOUBLE PRECISION,
            ticks INTEGER,
            PRIMARY KEY (pkey, dt),
            CONSTRAINT fk_pkey FOREIGN KEY (pkey) REFERENCES {ref_schema_name}.{ref_table_name} (pkey)
        );
        SELECT create_hypertable({full_name}, by_range('dt'));
    """
    ).format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(repr(table)),
        full_name=sql.Literal(schema + "." + repr(table)),
        # No easy way to make the pkey reference variable
        ref_schema_name=sql.Identifier(Schema.SECURITY),
        ref_table_name=sql.Identifier(AssetTbls.SYMBOLS),
    )


def create_continuous_aggrigate(schema: str, table: AssetTable, ref_table: AssetTable) -> sql.Composed:
    "Create a Higher-Timeframe Aggregate from an OHLC Dataset"
    _error_check_continuous_aggrigate(table, ref_table)
    return sql.SQL(
        """
        CREATE MATERIALIZED VIEW {schema_name}.{table_name} WITH (timescaledb.continuous) AS
        SELECT 
            time_bucket({interval}, dt, TIMESTAMPTZ {origin}) as dt,
            pkey,"""
        + ("    first(rth, dt) AS rth," if table.has_rth else "")
        + """
            first(open, dt) AS open,
            last(close, dt) AS close,
            max(high) AS high,
            min(low) AS low,
            sum(volume) AS volume,
            sum(vwap * volume) / NULLIF(SUM(volume), 0) AS vwap,
            sum(ticks) AS ticks
        FROM {schema_name}.{ref_table_name}"""
        # Where clause handles the case when going from a table with an rth column to no rth column
        + (" WHERE rth = 0 " if ref_table.rth is None and table.ext and table.rth else "")
        # GROUP By 1 == Group by dt. Must use number since there is a name conflict on 'dt'
        # between source table and the selected table. Names must be Identical to chain aggregates.
        + """
        GROUP BY pkey, 1 ORDER BY 1
        WITH NO DATA;
    """
    ).format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(repr(table)),
        origin=sql.Literal(table.origin),
        interval=sql.Literal(table.psql_interval),
        ref_table_name=sql.Identifier(repr(ref_table)),
    )


def create_raw_agg_buffer(table: AssetTable) -> sql.Composed:
    return sql.SQL(
        """
        CREATE TEMP TABLE _aggregate_buffer (
            dt TIMESTAMPTZ NOT NULL,"""
        + (" rth SMALLINT NOT NULL," if table.has_rth else "")
        + """
            close DOUBLE PRECISION NOT NULL,
            open DOUBLE PRECISION DEFAULT NULL,
            high DOUBLE PRECISION DEFAULT NULL,
            low DOUBLE PRECISION DEFAULT NULL,
            volume DOUBLE PRECISION DEFAULT NULL,
            vwap DOUBLE PRECISION DEFAULT NULL,
            ticks INTEGER DEFAULT NULL
        ) ON COMMIT DROP;
    """
    ).format()


def copy_aggregates(args: list[str]) -> sql.Composed:
    return sql.SQL("COPY _aggregate_buffer ({args}) FROM STDIN;").format(
        args=sql.SQL(",").join([sql.Identifier(arg) for arg in args]),
    )


def insert_copied_aggregates(schema: str, table: AssetTable, pkey: int) -> sql.Composed:
    if table.has_rth:
        return sql.SQL(
            """
            INSERT INTO {schema_name}.{table_name} (pkey, dt, rth, close, open, high, low, volume, vwap, ticks) 
            SELECT {pkey}, dt, rth, close, open, high, low, volume, vwap, ticks FROM _aggregate_buffer;
        """
        ).format(
            schema_name=sql.Identifier(schema),
            table_name=sql.Identifier(str(table)),
            pkey=sql.Literal(pkey),
        )
    else:
        return sql.SQL(
            """
            INSERT INTO {schema_name}.{table_name} (pkey, dt, close, open, high, low, volume, vwap, ticks) 
            SELECT {pkey}, dt, close, open, high, low, volume, vwap, ticks FROM _aggregate_buffer;
        """
        ).format(
            schema_name=sql.Identifier(schema),
            table_name=sql.Identifier(str(table)),
            pkey=sql.Literal(pkey),
        )


def upsert_copied_aggregates(schema: str, table: AssetTable, pkey: int) -> sql.Composed:
    "Not Intended to be used as often as INSERT Operation since this requires a CONTINUOUS AGG REFRESH"
    if table.has_rth:
        return sql.SQL(
            """
            INSERT INTO {schema_name}.{table_name} (pkey, dt, rth, close, open, high, low, volume, vwap, ticks) 
            SELECT {pkey}, dt, rth, close, open, high, low, volume, vwap, ticks FROM _aggregate_buffer
            ON CONFLICT (pkey, dt) DO UPDATE
            SET rth = EXCLUDED.rth,
                close = EXCLUDED.close,
                open = EXCLUDED.open, 
                high = EXCLUDED.high, 
                low = EXCLUDED.low, 
                volume = EXCLUDED.volume, 
                vwap = EXCLUDED.vwap, 
                ticks = EXCLUDED.ticks;
        """
        ).format(
            schema_name=sql.Identifier(schema),
            table_name=sql.Identifier(str(table)),
            pkey=sql.Literal(pkey),
        )
    else:
        return sql.SQL(
            """
            INSERT INTO {schema_name}.{table_name} (pkey, dt, close, open, high, low, volume, vwap, ticks) 
            SELECT {pkey}, dt, close, open, high, low, volume, vwap, ticks FROM _aggregate_buffer
            ON CONFLICT (pkey, dt) DO UPDATE
            SET close = EXCLUDED.close,
                open = EXCLUDED.open, 
                high = EXCLUDED.high, 
                low = EXCLUDED.low, 
                volume = EXCLUDED.volume, 
                vwap = EXCLUDED.vwap, 
                ticks = EXCLUDED.ticks;
        """
        ).format(
            schema_name=sql.Identifier(schema),
            table_name=sql.Identifier(str(table)),
            pkey=sql.Literal(pkey),
        )


def _array_select(column: str):
    return sql.SQL("ARRAY( SELECT {col} FROM inner_select) AS {col}").format(col=sql.Identifier(column))


def select_aggregates(
    schema: Schema,
    table: AssetTable,
    pkey: int,
    rth: bool,
    start: Optional[Timestamp],
    end: Optional[Timestamp],
    _limit: Optional[int],
    rtn_args: set[str],
) -> sql.Composed:

    _filters: list[Filter] = [("pkey", "=", pkey)]
    if start is not None:
        _filters.append(("dt", ">=", start))
    if end is not None:
        _filters.append(("dt", "<", end))

    # Filter by rth if accessing a table that has both rth and eth
    if rth and table.ext and table.rth is None:
        _filters.append(("rth", "=", 0))

    if "rth" in rtn_args and table.ext and table.rth is True:
        # 'rth' Doesn't exist in the table we are selecting from.
        rtn_args.remove("rth")

    rtn_args |= {"dt"}  # Ensure dt is returned
    if table.period == Timedelta(0):
        _ordered_rtn_args = [v for v in get_args(TickArgs) if v in rtn_args]
    else:
        _ordered_rtn_args = [v for v in get_args(AggregateArgs) if v in rtn_args]

    # Select all the needed data, then reorient it so it is returned by column instead of by row
    return sql.SQL(
        """
        WITH inner_select AS (
            SELECT {rtn_args} FROM {schema_name}.{table_name}{filter}{order}{limit}
        )
        SELECT {rtn_arrays} ;
    """
    ).format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(str(table)),
        rtn_args=arg_list(_ordered_rtn_args),
        filter=where(_filters),
        order=order("dt", True),
        limit=limit(_limit),
        rtn_arrays=sql.SQL(",\n").join([_array_select(col) for col in _ordered_rtn_args]),
    )


def select_aggregates_copy(
    mdata: MetadataInfo,
    rth: bool,
    start: Optional[Timestamp],
    end: Optional[Timestamp],
    _limit: Optional[int],
    rtn_args: set[str],
) -> sql.Composed:
    assert mdata.table

    _filters: list[Filter] = [("pkey", "=", mdata.pkey)]
    if start is not None:
        _filters.append(("dt", ">=", start))
    if end is not None:
        _filters.append(("dt", "<", end))

    # Filter by rth if accessing a table that has both rth and eth
    if rth and mdata.table.has_rth:
        _filters.append(("rth", "=", 0))

    if "rth" in rtn_args and not mdata.table.has_rth:
        # 'rth' Doesn't exist in the table we are selecting from.
        rtn_args.remove("rth")

    rtn_args -= {"dt"}  # dt already ensured as a return
    if mdata.table.period == Timedelta(0):
        dt_arg = "dt"
        _ordered_rtn_args = [v for v in get_args(TickArgs) if v in rtn_args]
    else:
        dt_arg = "EXTRACT(EPOCH FROM dt)::BIGINT as dt"
        _ordered_rtn_args = [v for v in get_args(AggregateArgs) if v in rtn_args]

    # Select all the needed data, then reorient it so it is returned by column instead of by row
    return sql.SQL(
        "COPY ( SELECT "
        + dt_arg
        + """, {rtn_args} FROM {schema_name}.{table_name}{filter}{order}{limit}
        ) TO STDOUT WITH CSV HEADER;
    """
    ).format(
        schema_name=sql.Identifier(mdata.schema_name),
        table_name=sql.Identifier(str(mdata.table)),
        rtn_args=arg_list(_ordered_rtn_args),
        filter=where(_filters),
        order=order("dt", True),
        limit=limit(_limit),
    )


def calculate_aggregates(
    schema: Schema,
    src_table: AssetTable,
    timeframe: Timedelta,
    pkey: int,
    rth: bool,
    start: Optional[Timestamp],
    end: Optional[Timestamp],
    _limit: Optional[int],
    rtn_args: set[str],
) -> sql.Composed:

    _filters: list[Filter] = [("pkey", "=", pkey)]
    if start is not None:
        _filters.append(("dt", ">=", start))
    if end is not None:
        _filters.append(("dt", "<", end))

    if src_table.has_rth and rth:
        # Filter by rth if accessing a table that has both rth and eth
        _filters.append(("rth", "=", 0))
    if not src_table.has_rth and "rth" in rtn_args:
        # 'rth' Doesn't exist in the table we are selecting from.
        rtn_args.remove("rth")

    if src_table.period == Timedelta(0):
        _inner_sel_args = _tick_inner_select_args(rtn_args)
    else:
        _inner_sel_args = _agg_inner_select_args(rtn_args)

    rtn_args |= {"dt"}  # Ensure dt is returned
    _ordered_rtn_args = [v for v in get_args(AggregateArgs) if v in rtn_args]

    return sql.SQL(
        """
        WITH inner_select AS (
            SELECT 
                time_bucket({interval}, dt, ({origin_select})) as dt,
                {inner_select_args}
            FROM {schema}.{table_name}{filters} GROUP BY 1 ORDER BY 1{limit} 
        )
        SELECT {rtn_arrays} ;
        """
    ).format(
        schema=sql.Identifier(schema),
        table_name=sql.Identifier(repr(src_table)),
        origin_select=_select_origin(schema, src_table.asset_class, rth, timeframe),
        interval=sql.Literal(str(int(timeframe.total_seconds())) + " seconds"),
        inner_select_args=sql.SQL(", ").join(_inner_sel_args),
        filters=where(_filters),
        limit=limit(_limit),
        rtn_arrays=sql.SQL(",\n").join([_array_select(col) for col in _ordered_rtn_args]),
    )


def calculate_aggregates_copy(
    mdata: MetadataInfo,
    timeframe: Timedelta,
    rth: bool,
    start: Optional[Timestamp],
    end: Optional[Timestamp],
    _limit: Optional[int],
    rtn_args: set[str],
) -> sql.Composed:
    assert mdata.table

    _filters: list[Filter] = [("pkey", "=", mdata.pkey)]
    if start is not None:
        _filters.append(("dt", ">=", start))
    if end is not None:
        _filters.append(("dt", "<", end))

    if mdata.table.has_rth and rth:
        # Filter by rth if accessing a table that has both rth and eth
        _filters.append(("rth", "=", 0))
    if not mdata.table.has_rth and "rth" in rtn_args:
        # 'rth' Doesn't exist in the table we are selecting from.
        rtn_args.remove("rth")

    rtn_args -= {"dt"}  # dt is Guaranteed to be returned
    if mdata.timeframe == Timedelta(0):
        _inner_sel_args = _tick_inner_select_args(rtn_args)
    else:
        _inner_sel_args = _agg_inner_select_args(rtn_args)

    return sql.SQL(
        """
        COPY (
            SELECT
                EXTRACT(EPOCH FROM (time_bucket({interval}, dt, ({origin_select}))))::BIGINT as dt,
                {inner_select_args}
            FROM {schema}.{table_name}{filters} GROUP BY 1 ORDER BY 1{limit} 
        ) TO STDOUT WITH CSV HEADER;
        """
    ).format(
        schema=sql.Identifier(mdata.schema_name),
        table_name=sql.Identifier(repr(mdata.table)),
        origin_select=_select_origin(mdata.schema_name, mdata.table.asset_class, rth, timeframe),
        interval=sql.Literal(str(int(timeframe.total_seconds())) + " seconds"),
        inner_select_args=sql.SQL(", ").join(_inner_sel_args),
        filters=where(_filters),
        limit=limit(_limit),
    )


def _agg_inner_select_args(args: set[str]) -> list[sql.Composable]:
    inner_select_args = []
    if "rth" in args:
        inner_select_args.append(sql.SQL("first(rth, dt) AS rth"))
    if "open" in args:
        inner_select_args.append(sql.SQL("first(open, dt) AS open"))
    if "high" in args:
        inner_select_args.append(sql.SQL("max(high) AS high"))
    if "low" in args:
        inner_select_args.append(sql.SQL("min(low) AS low"))
    if "close" in args:
        inner_select_args.append(sql.SQL("last(close, dt) AS close"))
    if "volume" in args:
        inner_select_args.append(sql.SQL("sum(volume) AS volume"))
    if "ticks" in args:
        inner_select_args.append(sql.SQL("sum(ticks) AS ticks"))
    if "vwap" in args:
        inner_select_args.append(sql.SQL("sum(vwap * volume) / NULLIF(SUM(volume), 0) AS vwap"))
    return inner_select_args


def _tick_inner_select_args(args: set[str]) -> list[sql.Composable]:
    inner_select_args = []
    if "rth" in args:
        inner_select_args.append(sql.SQL("first(rth, dt) AS rth"))
    if "open" in args:
        inner_select_args.append(sql.SQL("first(price, dt) AS open"))
    if "high" in args:
        inner_select_args.append(sql.SQL("max(price) AS high"))
    if "low" in args:
        inner_select_args.append(sql.SQL("min(price) AS low"))
    if "close" in args:
        inner_select_args.append(sql.SQL("last(price, dt) AS close"))
    if "volume" in args:
        inner_select_args.append(sql.SQL("sum(volume) AS volume"))
    if "ticks" in args:
        inner_select_args.append(sql.SQL("count(*) AS ticks"))
    if "vwap" in args:
        inner_select_args.append(sql.SQL("sum(price * volume) / NULLIF(SUM(volume), 0) AS vwap"))
    return inner_select_args


# endregion



"Psyscale Partial Class to add Timeseries Schema Configuration script"

from abc import abstractmethod
import logging
from typing import Any, Dict, Iterable, Literal, Optional, get_args
from itertools import chain

import psycopg as pg
from pandas import DataFrame, Timestamp, Timedelta

from psyscale.async_core import PsyscaleAsyncCore
from psyscale.psql.orm import MetadataArgs, MetadataInfo
from psyscale.psql.timeseries import AggregateArgs, TickArgs

from .psql import (
    GenericTbls,
    Operation as Op,
    Schema,
    SeriesTbls,
    TimeseriesConfig,
    Commands,
)

from .core import PsyscaleConnectParams, PsyscaleCore, TupleCursor

log = logging.getLogger("psyscale_log")

# pylint: disable='missing-function-docstring','unused-argument','protected-access'


class TimeseriesPartialAbstract(PsyscaleCore):
    """
    Psyscale Partial Timeseries Abstract base Class to add Timeseries Schema Configuration script.

    Defines abstract methods for MetadataPartial & Series Partial. Complicated Class structiure is
    done to organize the complex functionality required by the Psyscale Class.
    """

    def __init__(
        self,
        conn_params: Optional[PsyscaleConnectParams | str] = None,
        *,
        down_on_del: bool = False,
        docker_compose_fpath: Optional[str] = None,
    ):
        # Chain the __init__ Docstring up the MRO since nothing changed
        self.__class__.__init__.__doc__ = super().__init__.__doc__
        super().__init__(
            conn_params,
            down_on_del=down_on_del,
            docker_compose_fpath=docker_compose_fpath,
        )
        self._read_db_timeseries_config()

    def configure_timeseries_schema(
        self,
        tick_tables: Optional[TimeseriesConfig] = None,
        minute_tables: Optional[TimeseriesConfig] = None,
        aggregate_tables: Optional[TimeseriesConfig] = None,
    ):
        """
        Compare the given TimeseriesConfig Object to what is stored. Makes Database changes via CLI
        as needed. Deletion of Calculated and Stored information will be confirmed before execution.
        """

        with self._cursor() as cursor:
            cursor.execute(self[Op.SELECT, GenericTbls.SCHEMA]())
            schemas = set(get_args(Schema))

            # Check & Create Schemas
            for schema in schemas.difference({rsp[0] for rsp in cursor.fetchall()}):
                log.info("Creating Schema '%s'", schema)
                cursor.execute(self[Op.CREATE, GenericTbls.SCHEMA](schema))

            cursor.connection.commit()

            # Create Each Class of Timeseries Table
            if tick_tables is not None:
                _configure_timeseries_schema(self, cursor, Schema.TICK_DATA, tick_tables)
            if minute_tables is not None:
                _configure_timeseries_schema(self, cursor, Schema.MINUTE_DATA, minute_tables)
            if aggregate_tables is not None:
                _configure_timeseries_schema(self, cursor, Schema.AGGREGATE_DATA, aggregate_tables)

        # Ensure The appropriate timeseries info is stored in the event this class is used
        # directly after setting the config
        self._read_db_timeseries_config()

    def _read_db_timeseries_config(self):
        "Read off the TimeseriesConfig for each schema by probing all the table names."
        self._table_config: Dict[Schema, TimeseriesConfig] = {}

        with self._cursor() as cursor:
            for schema in (Schema.TICK_DATA, Schema.MINUTE_DATA, Schema.AGGREGATE_DATA):
                # ---- ---- Read the Origin Timestamp Table ---- ----
                origin_map = {}
                try:
                    cursor.execute(self[Op.SELECT, SeriesTbls._ORIGIN](schema, _all=True))
                    for (asset, *origins), *_ in cursor.fetchall():
                        # Cursed parsing for the cursor response tuple.
                        # Origins must be RTH, ETH, then HTF
                        origin_map[asset] = tuple(map(Timestamp, origins))

                except pg.DatabaseError:
                    # Origin Table does not exist, Rollback to clear error state
                    cursor.connection.rollback()
                    log.debug("Origin table not found in Schema: %s", schema)

                # ---- Reconstruct Timeseries Config from existing table names ----
                cursor.execute(self[Op.SELECT, GenericTbls.SCHEMA_TABLES](schema))
                tbl_names = [rsp[0] for rsp in cursor.fetchall() if rsp[0] != SeriesTbls._ORIGIN.value]
                config = TimeseriesConfig.from_table_names(tbl_names, origin_map)
                self._table_config[schema] = config

                # ---- ---- Check that all the origin times are preset ---- ----
                missing_asset_origins = set(config.asset_classes).difference(origin_map.keys())
                if len(missing_asset_origins) > 0:
                    log.error(
                        "TimescaleDB Origins Table in schema '%s' is missing values for the following assets: %s",
                        schema,
                        missing_asset_origins,
                    )

        # Give a notification on how to setup the database if it appears like it hasn't been
        all_assets = {chain(map(lambda x: x.asset_classes, self._table_config.values()))}
        if len(all_assets) == 0:
            log.warning(
                "No Asset Types Detected in the Database. To Initialize the Database call "
                "PsyscaleDB.configure_timeseries_schema() with the appropriate arguments.\n"
            )

    # region ---- ---- Abstract methods implemented by partial subclasses ---- ----
    @abstractmethod
    def inferred_metadata(self, symbol: int | str, timeframe: Timedelta, rth: bool = False) -> MetadataInfo | None: ...

    @abstractmethod
    def stored_metadata(
        self,
        symbol: int | str,
        filters: dict[MetadataArgs | str, Any] = {},
        *,
        _all: bool = False,
    ) -> list[MetadataInfo]: ...

    @abstractmethod
    def manually_refresh_aggregate_metadata(self): ...

    @abstractmethod
    def get_series(
        self,
        symbol: int | str,
        timeframe: Timedelta,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        limit: Optional[int] = None,
        rth: bool = False,
        rtn_args: Optional[Iterable[AggregateArgs | TickArgs | str]] = None,
        *,
        mdata: Optional[MetadataInfo] = None,
    ) -> Optional[DataFrame]: ...

    @abstractmethod
    def upsert_series(
        self,
        pkey: int,
        metadata: MetadataInfo,
        data: DataFrame,
        exchange: Optional[str] = None,
        *,
        on_conflict: Literal["update", "error", "ignore"] = "ignore",
    ): ...

    @abstractmethod
    def refresh_aggregate_metadata(self): ...

    # endregion


class TimeseriesPartialAsyncAbstract(PsyscaleAsyncCore):
    "Abstract Class for Async Timeseries Database Operations"

    @abstractmethod
    async def upsert_series_async(
        self,
        pkey: int,
        metadata: MetadataInfo,
        data: DataFrame,
        exchange: Optional[str] = None,
        *,
        on_conflict: Literal["update", "error", "ignore"] = "ignore",
    ): ...

    @abstractmethod
    async def get_series_async(
        self,
        symbol: int | str,
        timeframe: Timedelta,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        limit: Optional[int] = None,
        rth: bool = False,
        rtn_args: Optional[Iterable[AggregateArgs | TickArgs | str]] = None,
        *,
        mdata: Optional[MetadataInfo] = None,
    ) -> Optional[DataFrame]: ...

    @abstractmethod
    async def refresh_aggregate_metadata_async(self): ...


def _configure_timeseries_schema(
    db: TimeseriesPartialAbstract,
    cursor: TupleCursor,
    schema: Schema,
    config: TimeseriesConfig,
):
    "Script to Make Changes to the configuration of stored Timeseries Data"
    cursor.execute(db[Op.SELECT, GenericTbls.SCHEMA_TABLES](schema))
    tables: set[str] = {rsp[0] for rsp in cursor.fetchall()}
    log.info("---- ---- ---- Configuring Timeseries Schema '%s' ---- ---- ----", schema)

    # Ensure Origins Timestamp Table exists in the schema
    if SeriesTbls._ORIGIN not in tables:
        log.info("Creating '%s'.'%s' Table\n", schema, SeriesTbls._ORIGIN)
        cursor.execute(db[Op.CREATE, SeriesTbls._ORIGIN](schema))
    else:
        tables -= {SeriesTbls._ORIGIN.value}
        log.debug("'%s'.'%s' Table Already Exists\n", schema, SeriesTbls._ORIGIN)

    stored_config = db._table_config[schema]

    log.info("---- Checking for Assets that need to be added. ----")
    _add_timeseries_asset_classes(db.cmds, cursor, schema, config, stored_config)
    cursor.connection.commit()

    log.info("---- Checking for Assets that need to be Changed. ----")
    _update_timeseries_asset_classes(db.cmds, cursor, schema, config, stored_config)
    cursor.connection.commit()

    log.info("---- Checking for Assets that need to be Removed. ----")
    _del_timeseries_asset_classes(db.cmds, cursor, schema, config, stored_config)
    cursor.connection.commit()


def _add_timeseries_asset_classes(
    cmds: Commands,
    cursor: TupleCursor,
    schema: Schema,
    config: TimeseriesConfig,
    stored_config: TimeseriesConfig,
):
    additions = set(config.asset_classes).difference(stored_config.asset_classes)
    if len(additions) == 0:
        log.info("No Asset_classes need to be Added.")
        return

    for asset in additions:
        log.info("Generating all tables for asset_class: %s", asset)

        origin_args = {
            "rth_origin": config.rth_origins[asset],
            "eth_origin": config.eth_origins[asset],
            "htf_origin": config.htf_origins[asset],
        }
        log.info("Inserting Origin Timestamps: %s", origin_args)
        cursor.execute(cmds[Op.INSERT, SeriesTbls._ORIGIN](schema, asset, **origin_args))

        # Generate Raw insertion tables
        for tbl in config.raw_tables(asset):
            log.info("Generating table for: '%s'.'%s'", schema, tbl)
            tbl_type = SeriesTbls.TICK if tbl.period == Timedelta(0) else SeriesTbls.RAW_AGGREGATE

            cursor.execute(cmds[Op.CREATE, tbl_type](schema, tbl))

        # Generate Continuous Aggregates
        tbls = config.all_tables(asset, include_raw=False)
        tbls.sort(key=lambda x: x.period)  # Must generate lowest periods first
        for tbl in tbls:
            log.info("Generating Continuous Aggregate for: '%s'.'%s'", schema, tbl)
            ref_table = config.get_aggregation_source(tbl)
            tbl_type = SeriesTbls.CONTINUOUS_TICK_AGG if ref_table.period == Timedelta(0) else SeriesTbls.CONTINUOUS_AGG
            cursor.execute(cmd := cmds[Op.CREATE, tbl_type](schema, tbl, ref_table))
            log.debug("CMD: %s", cmd.as_string())


def _update_timeseries_asset_classes(
    cmds: Commands,
    cursor: TupleCursor,
    schema: Schema,
    config: TimeseriesConfig,
    stored_config: TimeseriesConfig,
):
    asset_updates = set(config.asset_classes).intersection(stored_config.asset_classes)
    if len(asset_updates) == 0:
        log.info("No Asset_classes need to be Updated.")
        return

    for asset in asset_updates:
        removals = set(stored_config.all_tables(asset)).difference(config.all_tables(asset))

        additions = set(config.all_tables(asset)).difference(stored_config.all_tables(asset))

        if len(removals) == 0 and len(additions) == 0:
            log.info("No changes needed for asset_class: %s", asset)
            continue

        _del = input(
            f"Aggregated Data Table Changes exist for Asset_class: '{schema}'.'{asset}'\n"
            "Updating these changes requires all Calculated Aggregates to be removed and "
            "recalculated.\n-- All Inserted data *will* be retained --\n"
            "Update Config? y/[N] : "
        )
        if not (_del == "y" or _del == "Y"):
            continue

        origin_args = {
            "rth_origin": config.rth_origins[asset],
            "eth_origin": config.eth_origins[asset],
            "htf_origin": config.htf_origins[asset],
        }
        log.info("Updating Origin Timestamps: %s", origin_args)
        cursor.execute(cmds[Op.UPDATE, SeriesTbls._ORIGIN](schema, asset, **origin_args))

        # Remove All Calculated Data Tables
        log.info("Updating config for Asset: %s", asset)

        # Must Remove Longest Aggregates first.
        all_aggregates = stored_config.all_tables(asset, include_raw=False)
        all_aggregates.sort(key=lambda x: x.period, reverse=True)
        for tbl in all_aggregates:
            log.info("Dropping Table: %s", tbl.table_name)
            cursor.execute(cmds[Op.DROP, GenericTbls.VIEW](schema, tbl.table_name))

        # Remove Unwanted Inserted Table Data
        for tbl in [tbl for tbl in removals if tbl.raw]:
            _del = input(
                f"Inserted Data Table '{tbl}' exists in current database, but not in the new "
                "config.\nThis table contains inserted raw data with an aggregation period of "
                f"{tbl.period}. \nDelete it? y/[N] : "
            )
            # Technically this introduces a bug but it's too much an edge case to care atm.
            # If the table is retained it will only be used for data retrieval after restart.
            # Despite if it is the lowest timeframe and should be used as the source for all
            # aggregations.
            if not (_del == "y" or _del == "Y"):
                continue

            log.info("Dropping Inserted Table: %s", tbl.table_name)
            cursor.execute(cmds[Op.DROP, GenericTbls.TABLE](schema, tbl.table_name))

        # Create new Raw Tables
        for tbl in [tbl for tbl in additions if tbl.raw]:
            log.info("Generating table for: '%s'.'%s'", schema, tbl)
            tbl_type = SeriesTbls.TICK if tbl.period == Timedelta(0) else SeriesTbls.RAW_AGGREGATE

            cursor.execute(cmds[Op.CREATE, tbl_type](schema, tbl))

        # Generate Continuous Aggregates
        tbls = config.all_tables(asset, include_raw=False)
        tbls.sort(key=lambda x: x.period)  # Must generate lowest periods first
        for tbl in tbls:
            log.info("Generating Continuous Aggregate for: '%s'.'%s'", schema, tbl)
            ref_table = config.get_aggregation_source(tbl)
            tbl_type = SeriesTbls.CONTINUOUS_TICK_AGG if ref_table.period == Timedelta(0) else SeriesTbls.CONTINUOUS_AGG
            cursor.execute(cmd := cmds[Op.CREATE, tbl_type](schema, tbl, ref_table))
            log.debug("CMD: %s", cmd.as_string())


def _del_timeseries_asset_classes(
    cmds: Commands,
    cursor: TupleCursor,
    schema: Schema,
    config: TimeseriesConfig,
    stored_config: TimeseriesConfig,
):

    removals = set(stored_config.asset_classes).difference(config.asset_classes)
    if len(removals) == 0:
        log.info("No Asset_classes need to be removed.")
        return

    for asset in removals:
        log.info("Checking if asset_class should be removed: %s", asset)

        _del = input(
            f"Asset_class: '{schema}'.'{asset}' exists in current database, "
            "but not in the given config. Remove it? y/[N] : "
        )

        if not (_del == "y" or _del == "Y"):
            log.info("Keeping asset_class: %s", asset)
            continue

        _del = input("This will permanently remove all Downloaded and Calculated Data. Are you Sure? y/[N] : ")

        if not (_del == "y" or _del == "Y"):
            log.info("Keeping asset_class: %s", asset)
            continue

        log.info("Removing Asset Class: %s", asset)
        cursor.execute(cmds[Op.DELETE, SeriesTbls._ORIGIN](schema, asset))

        # Must delete Largest Aggregates First
        tbls = stored_config.all_tables(asset)
        tbls.sort(key=lambda x: x.period, reverse=True)
        for tbl in tbls:
            # Catch all Table Type for Generic Drop Commands, Will Cascade
            tbl_type = GenericTbls.TABLE if tbl.raw else GenericTbls.VIEW
            cursor.execute(cmd := cmds[Op.DROP, tbl_type](schema, tbl.table_name))
            log.debug(cmd.as_string())


# pylint: disable='wrong-import-position'
# At EOF import the Partials that are built on the above abstract classes
# to create the partials that have the full functionality and are actually used.
from .metadata_partial import MetadataPartial
from .series_data_partial import SeriesDataPartial, AsyncSeriesDataPartial


class TimerseriesPartial(MetadataPartial, SeriesDataPartial):
    "Partial PsyscaleDB Class that includes all SeriesData & Metadata Functions"


class TimeseriesAsyncPartial(MetadataPartial, AsyncSeriesDataPartial):
    "Partial PsyscaleAsync Class that includes all Sync & Async SeriesData & Metadata Functions"



from enum import Enum, StrEnum, auto


class Operation(Enum):
    "Postgres Operations"

    CREATE = auto()
    INSERT = auto()
    UPSERT = auto()
    UPDATE = auto()
    SELECT = auto()
    COPY = auto()
    DROP = auto()
    DELETE = auto()
    REFRESH = auto()


class Schema(StrEnum):
    "Schema Names available within the database"

    SECURITY = auto()
    USER_DATA = auto()
    TICK_DATA = auto()
    MINUTE_DATA = auto()
    AGGREGATE_DATA = auto()


class GenericTbls(StrEnum):
    "Generic Cmds that may apply to multiple table types"

    SCHEMA = auto()
    SCHEMA_TABLES = auto()
    TABLE = auto()
    VIEW = auto()


class SeriesTbls(StrEnum):
    "Raw & Aggregated Timeseries Data Tables"

    _ORIGIN = auto()
    TICK = auto()
    TICK_BUFFER = auto()
    RAW_AGGREGATE = auto()
    RAW_AGG_BUFFER = auto()
    CONTINUOUS_AGG = auto()
    CONTINUOUS_TICK_AGG = auto()
    # Following is Not a stored_table, just references a specific select function
    CALCULATE_AGGREGATE = auto()


class AssetTbls(StrEnum):
    "Security Information Tables"

    _METADATA = auto()
    _METADATA_FUNC = auto()
    SYMBOLS = auto()
    SYMBOLS_BUFFER = auto()
    _SYMBOL_SEARCH_FUNCS = auto()

    # Not Implemented Yet. All Data is assumed to be Adjusted.
    # SPLITS = auto()
    # DIVIDENDS = auto()
    # EARNINGS = auto()



"""SQL Operation Map Construction and Commands Accessor"""

from __future__ import annotations
from copy import deepcopy
from enum import StrEnum
from logging import getLogger
from functools import partial
from typing import Optional, Tuple, TypeAlias, Callable

from psycopg import sql

from . import generic as gen
from . import security as sec
from . import timeseries as ts
from .enum import Operation, Schema, GenericTbls, AssetTbls, SeriesTbls

log = getLogger("psyscale_log")


# pylint: disable=protected-access
class Commands:
    """
    Class that stores formattable Postgres Commands based on operation and table type.

    Extendable with custom PostgreSQL functions. Given functions will override base functions
    in the event a new function is given for a given operation, table pair.

    Since the Table Key is a StrEnum, overriding will occur if the StrEnums have identical values,
    even if they are separately defined StrEnums.
    """

    def __init__(self, operation_map: Optional[OperationMap] = None) -> None:
        self.operation_map = deepcopy(OPERATION_MAP)

        if operation_map is not None:
            self.merge_operations(operation_map)

    def merge_operations(self, operation_map: OperationMap):
        """
        Merge additional operations into the existing suite of pre-formatted SQL Commands.
        Merging prefers the operations given over the operations present.

        Commands are access via Operation & strEnums. strEnums are keyed as strings, not object ids.
        Therefore, overwriting not only occurs when a command is passed with an already existing
        [Operation, Table_type] but also when a command is passed as [Operation, str(table_type)]

        e.g.
        [Operation.Create, AssetTbls.Symbol] = sql.SQL() & [Operation.Create, 'symbol'] = sql.SQL()
        will both overwrite the creation of the symbols table.
        """
        for operation, tbl_map in operation_map.items():
            known_keys = set(self.operation_map[operation].keys())
            if len(overlap := known_keys.intersection(tbl_map.keys())) > 0:
                log.warning(
                    "Overriding psyscale default %s for tables: %s",
                    operation,
                    overlap,
                )
            self.operation_map[operation] |= tbl_map

    def __getitem__(self, args: Tuple[Operation, StrEnum | str]) -> Callable[..., sql.Composed]:
        """
        Accessor to retrieve sql commands. Does not type check function args.
        Call Signature is Obj[Operation, Table](*Function Specific args)
        """
        if args[1] not in self.operation_map[args[0]]:
            tbl = args[1] if isinstance(args[1], str) else args[1].value
            raise ValueError(f"Operation '{args[0].name}' not known for Postgres Table: {tbl}")

        return self.operation_map[args[0]][args[1]]


OperationMap: TypeAlias = dict[Operation, dict[StrEnum | str, Callable[..., sql.Composed]]]

OPERATION_MAP: OperationMap = {
    # Mapping that defines the SQL Composing Function for each Operation and Table Combination
    Operation.CREATE: {
        GenericTbls.SCHEMA: gen.create_schema,
        SeriesTbls._ORIGIN: ts.create_origin_table,
        SeriesTbls.TICK: ts.create_tick_table,
        SeriesTbls.TICK_BUFFER: ts.create_raw_tick_buffer,
        SeriesTbls.RAW_AGGREGATE: ts.create_raw_aggregate_table,
        SeriesTbls.RAW_AGG_BUFFER: ts.create_raw_agg_buffer,
        SeriesTbls.CONTINUOUS_AGG: ts.create_continuous_aggrigate,
        SeriesTbls.CONTINUOUS_TICK_AGG: ts.create_continuous_tick_aggregate,
        AssetTbls.SYMBOLS: sec.create_symbol_table,
        AssetTbls.SYMBOLS_BUFFER: sec.create_symbol_buffer,
        AssetTbls._SYMBOL_SEARCH_FUNCS: sec.create_search_functions,
        AssetTbls._METADATA: sec.create_timeseries_metadata_view,
        AssetTbls._METADATA_FUNC: sec.create_timeseries_metadata_subfunction,
    },
    Operation.INSERT: {
        SeriesTbls.TICK_BUFFER: ts.insert_copied_ticks,
        SeriesTbls.RAW_AGG_BUFFER: ts.insert_copied_aggregates,
        AssetTbls.SYMBOLS_BUFFER: sec.insert_copied_symbols,
        SeriesTbls._ORIGIN: ts.insert_origin,
    },
    Operation.UPSERT: {
        SeriesTbls.TICK_BUFFER: ts.upsert_copied_ticks,
        SeriesTbls.RAW_AGG_BUFFER: ts.upsert_copied_aggregates,
        AssetTbls.SYMBOLS_BUFFER: sec.upsert_copied_symbols,
    },
    Operation.UPDATE: {
        GenericTbls.TABLE: gen.update,
        SeriesTbls._ORIGIN: ts.update_origin,
        AssetTbls.SYMBOLS: partial(gen.update, Schema.SECURITY, AssetTbls.SYMBOLS),
    },
    Operation.COPY: {
        AssetTbls.SYMBOLS_BUFFER: sec.copy_symbols,
        SeriesTbls.TICK_BUFFER: ts.copy_ticks,
        SeriesTbls.RAW_AGG_BUFFER: ts.copy_aggregates,
        SeriesTbls.RAW_AGGREGATE: ts.select_aggregates_copy,
        SeriesTbls.CALCULATE_AGGREGATE: ts.calculate_aggregates_copy,
    },
    Operation.SELECT: {
        GenericTbls.TABLE: gen.select,
        GenericTbls.VIEW: gen.list_mat_views,
        GenericTbls.SCHEMA: gen.list_schemas,
        GenericTbls.SCHEMA_TABLES: gen.list_tables,
        SeriesTbls._ORIGIN: ts.select_origin,
        SeriesTbls.RAW_AGGREGATE: ts.select_aggregates,
        SeriesTbls.CALCULATE_AGGREGATE: ts.calculate_aggregates,
        AssetTbls.SYMBOLS: sec.select_symbols,
        AssetTbls._METADATA: sec.select_timeseries_metadata,
    },
    Operation.DELETE: {
        GenericTbls.TABLE: gen.delete,
        SeriesTbls._ORIGIN: ts.delete_origin,
    },
    Operation.DROP: {
        GenericTbls.SCHEMA: gen.drop_schema,
        GenericTbls.TABLE: gen.drop_table,
        GenericTbls.VIEW: gen.drop_materialized_view,
    },
    Operation.REFRESH: {
        AssetTbls._METADATA: sec.refresh_timeseries_metadata_view,
        SeriesTbls.CONTINUOUS_AGG: ts.refresh_continuous_aggregate,
        SeriesTbls.CONTINUOUS_TICK_AGG: ts.refresh_continuous_aggregate,
    },
}



"""
Wrapper Class for Timeseries Dataframes.
Class Standardizes Column Names & uses pandas_market_calendars
to determine the Trading Hours session.
"""

from importlib import import_module
import logging
from functools import partial
from types import ModuleType
from typing import TYPE_CHECKING, Dict, Optional

import pandas as pd


log = logging.getLogger("psyscale_log")

# Trading Hours Integer Encoding
EXT_MAP = {
    "pre": 1,
    "rth_pre_break": 0,
    "rth": 0,
    "break": 3,
    "rth_post_break": 0,
    "post": 2,
    "closed": -1,
}

# pylint: disable=line-too-long, invalid-name
# region ------------------------------ DataFrame Functions ------------------------------ #


def determine_timedelta(series: pd.DatetimeIndex | pd.Series) -> pd.Timedelta:
    "Returns the most frequent Timedelta within the first 250 indices of the data given"
    if isinstance(series, pd.DatetimeIndex):
        # .diff() Unknown-attribute False Alarm Error.
        return pd.Timedelta(series[0:250].diff().value_counts().idxmax())  # type: ignore
    else:
        return pd.Timedelta(series.iloc[0:250].diff().value_counts().idxmax())


def _standardize_names(df: pd.DataFrame):
    """
    Standardize the column names of the given dataframe to a consistent format for
    OHLC and Single Value Time-series. Changes are made inplace.
    """
    if isinstance(df.index, pd.DatetimeIndex):
        # In the event the timestamp is the index, reset it for naming
        df.reset_index(inplace=True, names="time")

    rename_map = {}
    df.columns = list(map(str.lower, df.columns))
    column_names = set(df.columns)

    # |= syntax merges the returned mapping into rename_map
    rename_map |= _column_name_check(
        column_names,
        ["dt", "t", "time", "date", "datetime", "timestamp"],
        True,
    )

    # These names are mostly chosen to match what Lightweight-Charts expects as input data
    rename_map |= _column_name_check(column_names, ["open", "o", "first"])
    rename_map |= _column_name_check(column_names, ["close", "c", "last"])
    rename_map |= _column_name_check(column_names, ["high", "h", "max"])
    rename_map |= _column_name_check(column_names, ["low", "l", "min"])
    rename_map |= _column_name_check(column_names, ["volume", "v", "vol"])
    rename_map |= _column_name_check(column_names, ["price", "val", "data", "value"])
    rename_map |= _column_name_check(column_names, ["vwap", "vw"])
    rename_map |= _column_name_check(column_names, ["ticks", "tick", "count", "trade_count", "n"])

    if len(rename_map) > 0:
        return df.rename(columns=rename_map, inplace=True)


def _column_name_check(
    column_names: set[str],
    aliases: list[str],
    required: bool = False,
) -> Dict[str, str]:
    """
    Checks the column names for any of the expected aliases.
    If required and not present, an Attribute Error is thrown.

    Returns a mapping of the {'aliases[0]': 'Found Alias'} if necessary
    """
    intersection = list(column_names.intersection(aliases))

    if len(intersection) == 0:
        if required:
            raise AttributeError(f'Given data must have a "{" | ".join(aliases)}" column')
        return {}

    if len(intersection) > 1:
        raise AttributeError(f'Given data can have only one "{" | ".join(aliases)}" type of column')

    return {intersection[0]: aliases[0]}


# endregion

# region --------------------------- Pandas Dataframe Object Wrappers --------------------------- #


class Series_DF:
    "Used to standardize column names and Mark EXT Session using pandas market calendars."

    def __init__(
        self,
        pandas_df: pd.DataFrame,
        exchange: Optional[str] = None,
    ):
        _standardize_names(pandas_df)
        # Set Consistent Time format (Pd.Timestamp, UTC, TZ Aware)
        pandas_df["dt"] = pd.to_datetime(pandas_df["dt"], utc=True)
        self.timedelta = determine_timedelta(pandas_df["dt"])
        self.calendar = CALENDARS.request_calendar(exchange, pandas_df["dt"].iloc[0], pandas_df["dt"].iloc[-1])

        self.df = pandas_df.set_index("dt", drop=False)
        self._mark_ext()

    # region --------- Properties --------- #

    @property
    def columns(self) -> set[str]:
        "Column Names within the Dataframe"
        return set(self.df.columns)

    @property
    def ext(self) -> bool | None:
        "True if data has Extended Trading Hours Data, False if no ETH Data, None if undefined."
        return self._ext

    @property
    def _dt_index(self) -> pd.DatetimeIndex:
        return pd.DatetimeIndex(self.df.index)  # type:ignore

    # endregion

    def _mark_ext(self, force_rth: bool = False):
        if "rth" in self.columns:
            # In case only part of the df has ext classification, fill the remainder
            missing_rth = self._dt_index[self.df["rth"].isna()]
            rth_col = CALENDARS.mark_session(self.calendar, missing_rth)
            if rth_col is not None:
                self.df.loc[rth_col.index, "rth"] = rth_col
        else:
            # Calculate the Full Trading Hours Session
            rth_col = CALENDARS.mark_session(self.calendar, self._dt_index)
            if rth_col is not None:
                self.df["rth"] = rth_col

        if "rth" not in self.columns:
            self._ext = None
        elif force_rth:
            self.df = self.df[self.df["rth"] == EXT_MAP["rth"]]
            self._ext = False
        elif (self.df["rth"] == 0).all():
            # Only RTH Sessions
            self._ext = False
        else:
            # Some RTH, Some ETH Sessions
            self._ext = True


# endregion

# region --------------------------- Pandas_Market_Calendars Adapter --------------------------- #

if TYPE_CHECKING:
    import pandas_market_calendars as mcal
    from pandas_market_calendars import MarketCalendar
else:
    mcal: Optional[ModuleType] = None


class Calendars:
    """
    Class abstracts and contains the functionality of pandas_market_calendars.

    Instantiating only a single instance reduces unnecessary redundancy by making
    market schedules shared across all dataframes that utilize them. Considering that generating
    schedules is easily the slowest part of analyzing a Market's Open/Close Session this equates
    to a significant performance improvement.
    """

    EXCHANGE_NAMES = {}
    ALT_EXCHANGE_NAMES = {
        "xnas": "NASDAQ",
        "arca": "NYSE",
        "forex": "24/5",
        "crypto": "24/7",
    }

    def __init__(self):
        self.mkt_cache: Dict[str, MarketCalendar] = {}
        self.schedule_cache: Dict[str, pd.DataFrame] = {}

    def _import_mcal(self):
        # Lazy load Pandas_market_calendars
        # pylint: disable-next='global-statement'
        global mcal
        mcal = import_module("pandas_market_calendars")
        self.EXCHANGE_NAMES = dict([(val.lower(), val) for val in mcal.get_calendar_names()])

    def request_calendar(self, exchange: Optional[str], start: pd.Timestamp, end: pd.Timestamp) -> str:
        "Request a Calendar & Schedule be Cached. Returns a token to access the cached calendar"
        if mcal is None:
            self._import_mcal()

        if exchange is None:
            return "24/7"
        exchange = exchange.lower()
        if exchange in self.ALT_EXCHANGE_NAMES:
            cal = mcal.get_calendar(self.ALT_EXCHANGE_NAMES[exchange])
        elif exchange in self.EXCHANGE_NAMES:
            cal = mcal.get_calendar(self.EXCHANGE_NAMES[exchange])
        else:
            cal = None
            log.warning(
                "Exchange '%s' doesn't match any exchanges. Using 24/7 Calendar.",
                exchange,
            )

        if cal is None or cal.name == "24/7":
            return "24/7"

        start = start - pd.Timedelta("1W")
        end = end + pd.Timedelta("1W")

        if cal.name not in self.mkt_cache:  # New Calendar Requested
            # Bind the Market_times & special_times arguments to the schedule function
            cal.schedule = partial(  # type:ignore
                cal.schedule, market_times="all", force_special_times=False
            )
            self.mkt_cache[cal.name] = cal
            # Generate a Schedule with buffer dates on either side.
            self.schedule_cache[cal.name] = cal.schedule(start, end)
            return cal.name

        # Cached Calendar Requested
        extra_dates = None
        cal = self.mkt_cache[cal.name]
        sched = self.schedule_cache[cal.name]
        if sched.index[0] > start.tz_localize(None):
            # Extend Start of Schedule with an additional buffer
            extra_dates = cal.schedule(start, sched.index[0] - pd.Timedelta("1D"))
            sched = pd.concat([extra_dates, sched])
        if sched.index[-1] < end.normalize().tz_localize(None):
            # Extend End of Schedule with an additional buffer
            extra_dates = cal.schedule(sched.index[-1] + pd.Timedelta("1D"), end)
            sched = pd.concat([sched, extra_dates])

        if extra_dates is not None:
            self.schedule_cache[cal.name] = sched

        return cal.name

    def mark_session(self, calendar: str, time_index: pd.DatetimeIndex) -> pd.Series | None:
        "Return a Series that denotes the appropriate Trading Hours Session for the given Calendar"
        if mcal is None or calendar == "24/7":
            return None

        return mcal.mark_session(self.schedule_cache[calendar], time_index, label_map=EXT_MAP, closed="left")


# Prep the shared Calendars sudo-singleton, will init when needed
CALENDARS = Calendars()

# endregion



import os
from unittest.mock import patch
from urllib.parse import quote_plus

import pytest
from psyscale import PsyscaleConnectParams


def test_direct_instantiation_sets_correct_url():
    params = PsyscaleConnectParams(
        host="localhost",
        port=5432,
        user="test_user",
        password="test_pass",
        database="test_db",
        sslmode="require",
        application_name="psyscale_test",
    )
    encoded_user = quote_plus("test_user")
    encoded_pass = quote_plus("test_pass")
    expected_url = (
        f"postgresql://{encoded_user}:{encoded_pass}@localhost:5432/test_db"
        f"?sslmode=require&application_name=psyscale_test"
    )
    assert params.url == expected_url
    assert params.is_local


def test_direct_instantiation_without_sslmode_or_app_name():
    params = PsyscaleConnectParams(host="127.0.0.1", port=5432, user="admin", password="admin123", database="db")
    assert "sslmode" not in params.url
    assert "application_name" not in params.url
    assert params.is_local


def test_from_url_parses_correctly():
    url = "postgresql://user:pass@myhost:5432/mydb?sslmode=disable&application_name=test"
    params = PsyscaleConnectParams.from_url(url)

    assert params.host == "myhost"
    assert params.port == 5432
    assert params.user == "user"
    assert params.password == "pass"
    assert params.database == "mydb"
    assert params.sslmode == "disable"
    assert params.application_name == "test"
    assert not params.is_local


def test_from_url_invalid_scheme_raises():
    with pytest.raises(ValueError, match="Invalid URL scheme"):
        PsyscaleConnectParams.from_url("mysql://user:pass@localhost:3306/db")


@patch.dict(
    os.environ,
    {
        "PSYSCALE_HOST": "localhost",
        "PSYSCALE_PORT": "5432",
        "PSYSCALE_USER": "testuser",
        "PSYSCALE_PASSWORD": "testpw",
        "PSYSCALE_DB_NAME": "testdb",
        "PSYSCALE_SSLMODE": "verify-full",
        "PSYSCALE_APP_NAME": "env_app",
        "PSYSCALE_VOLUME_PATH": "/data/db",
    },
    clear=True,
)
def test_from_env_without_url_uses_individual_vars():
    params = PsyscaleConnectParams.from_env()

    assert params.host == "localhost"
    assert params.port == 5432
    assert params.user == "testuser"
    assert params.password == "testpw"
    assert params.database == "testdb"
    assert params.sslmode == "verify-full"
    assert params.application_name == "env_app"
    assert params.volume_path == "/data/db"
    assert params.is_local


@patch.dict(
    os.environ,
    {"PSYSCALE_URL": "postgresql://u:p@host:1234/db?sslmode=require&application_name=env_url"},
    clear=True,
)
def test_from_env_with_url_parses_correctly():
    params = PsyscaleConnectParams.from_env()

    assert params.host == "host"
    assert params.port == 1234
    assert params.user == "u"
    assert params.password == "p"
    assert params.database == "db"
    assert params.sslmode == "require"
    assert params.application_name == "env_url"
    assert not params.volume_path
    assert not params.is_local


@pytest.mark.parametrize("host", ["localhost", "127.0.0.1", "::1"])
def test_is_local_true_for_local_hosts(host):
    params = PsyscaleConnectParams(host=host, port=5432, user="u", password="p", database="db")
    assert params.is_local


def test_is_local_false_for_remote_host():
    params = PsyscaleConnectParams(host="remote-db.example.com", port=5432, user="u", password="p", database="db")
    assert not params.is_local



"""Generic PostgreSQL Commands & Util Functions to Compose Filters"""

from enum import StrEnum
from typing import (
    Any,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypeAlias,
)

from psycopg import sql

ArgumentLiteral: TypeAlias = Tuple[str, Any]
Argument: TypeAlias = ArgumentLiteral | str
Comparators: TypeAlias = Literal["=", "!=", ">", "<", ">=", "<=", "LIKE", "ILIKE"]
FilterLiteral: TypeAlias = Tuple[str, Comparators, Any]
FilterPlaceholder: TypeAlias = Tuple[str, Comparators]
Filter: TypeAlias = sql.Composable | FilterLiteral | FilterPlaceholder

# region --------   -------- Filter Composer Utils  --------  --------
# pylint: disable=missing-function-docstring
# Util functions to add 'WHERE', 'LIMIT', and 'ORDER' Statements to a SQL Statement


def where(filters: Sequence[Filter] | Filter) -> sql.Composable:
    "SQL WHERE Clause to precede a set of filters"
    if isinstance(filters, (sql.Composable, Tuple)):
        return sql.SQL(" WHERE ") + filter_composer([filters])  # type:ignore
    else:
        return sql.SQL("") if len(filters) == 0 else sql.SQL(" WHERE ") + filter_composer(filters)


def limit(val: Optional[str | int]) -> sql.Composable:
    if val is None:
        return sql.SQL("")
    if isinstance(val, int):
        return sql.SQL(" LIMIT {val}").format(val=sql.Literal(val))
    else:
        return sql.SQL(" LIMIT {val_ph}").format(val_ph=sql.Placeholder(val))


def order(arg: Optional[str], ascending: Optional[bool] = True) -> sql.Composable:
    if arg is None:
        return sql.SQL("")
    if ascending is False:
        return sql.SQL(" ORDER BY {arg} DESC").format(arg=sql.Identifier(arg))
    else:
        return sql.SQL(" ORDER BY {arg}").format(arg=sql.Identifier(arg))


def arg_list(args: Sequence[str], distinct: bool = False) -> sql.Composable:
    fmt_args = sql.SQL("*") if len(args) == 0 else sql.SQL(", ").join(map(sql.Identifier, args))
    return sql.SQL("DISTINCT ") + fmt_args if distinct else fmt_args


def update_args(args: Sequence[Argument]) -> sql.Composable:
    if len(args) == 0:
        raise ValueError("Attempting to update arguments, but no values given to SET.")
    composables = [_arg_placeholder(v) if isinstance(v, str) else _arg_literal(*v) for v in args]
    return sql.SQL(", ").join(composables)


def _arg_placeholder(arg: str) -> sql.Composable:
    return sql.SQL("{arg} " + "=" + " {arg_ph}").format(
        arg=sql.Identifier(arg),
        arg_ph=sql.Placeholder(arg),
    )


def _arg_literal(arg: str, value: Any) -> sql.Composable:
    return sql.SQL("{arg} " + "=" + " {arg_lit}").format(
        arg=sql.Identifier(arg),
        arg_lit=sql.Literal(value),
    )


def _filter_literal(arg: str, comparison: Comparators, value: Any) -> sql.Composable:
    return sql.SQL("{arg}" + comparison + "{arg_val}").format(
        arg=sql.Identifier(arg),
        arg_val=sql.Literal(value),
    )


def _filter_placeholder(arg: str, comparison: Comparators) -> sql.Composable:
    return sql.SQL("{arg}" + comparison + "{arg_ph}").format(
        arg=sql.Identifier(arg),
        arg_ph=sql.Placeholder(arg),
    )


def filter_composer(filters: Sequence[Filter], mode: Literal["AND", "OR"] = "AND") -> sql.Composable:
    composables = [
        (
            sql.SQL("(") + v + sql.SQL(")")
            if isinstance(v, sql.Composable)
            else _filter_literal(*v) if len(v) == 3 else _filter_placeholder(*v)
        )
        for v in filters
    ]
    if mode == "OR":
        return sql.SQL(" OR ").join(composables)
    else:
        return sql.SQL(" AND ").join(composables)


# endregion


# region -------- -------- Generic PSQL Commands -------- --------


def list_schemas() -> sql.Composed:
    return sql.SQL("SELECT schema_name FROM information_schema.schemata;").format()


def list_mat_views(schema: str) -> sql.Composed:
    return sql.SQL("SELECT matviewname FROM pg_matviews WHERE schemaname = {schema_name};").format(
        schema_name=sql.Literal(schema),
    )


def list_tables(schema: str) -> sql.Composed:
    return sql.SQL("SELECT table_name FROM information_schema.tables WHERE table_schema = {schema_name};").format(
        schema_name=sql.Literal(schema),
    )


def create_schema(schema: str) -> sql.Composed:
    return sql.SQL("CREATE SCHEMA {schema_name};").format(
        schema_name=sql.Identifier(schema),
    )


def drop_schema(schema: str, cascade: bool = True) -> sql.Composed:
    return sql.SQL("DROP SCHEMA IF EXISTS {schema_name}" + " CASCADE" if cascade else "" + ";").format(
        schema_name=sql.Identifier(schema),
    )


def drop_table(schema: str, table: str, cascade: bool = True) -> sql.Composed:
    return sql.SQL("DROP TABLE IF EXISTS {schema_name}.{table_name}" + " CASCADE" if cascade else "" + ";").format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(table),
    )


def drop_materialized_view(schema: str, table: str, cascade: bool = True) -> sql.Composed:
    return sql.SQL(
        "DROP MATERIALIZED VIEW IF EXISTS {schema_name}.{table_name}" + " CASCADE" if cascade else "" + ";"
    ).format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(table),
    )


def delete(schema: str, table: str, filters: list[Filter], cascade: bool = True) -> sql.Composed:
    return sql.SQL("DELETE FROM {schema_name}.{table_name} {filter}" + " CASCADE" if cascade else "" + ";").format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(table),
        filter=where(filters),
    )


def select(
    schema: str | StrEnum,
    table: str | StrEnum,
    arguments: Sequence[str] = [],
    filters: Sequence[Filter] = [],
    _limit: Optional[str | int] = None,
    _order: Tuple[str | None, bool | None] = (None, None),
    *,
    distinct: bool = False,
) -> sql.Composed:
    return sql.SQL("SELECT {rtn_args} FROM {schema_name}.{table_name}{filter}{order}{limit};").format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(table),
        rtn_args=arg_list(arguments, distinct),
        filter=where(filters),
        order=order(*_order),
        limit=limit(_limit),
    )


def update(
    schema: str | StrEnum,
    table: str | StrEnum,
    assignments: list[Argument],
    filters: list[Filter],
) -> sql.Composed:
    return sql.SQL("UPDATE {schema_name}.{table_name} SET {args} {filter};").format(
        schema_name=sql.Identifier(schema),
        table_name=sql.Identifier(table),
        args=update_args(assignments),
        filter=where(filters),
    )


# endregion



import os
import pytest
from psycopg import connect
from psycopg.rows import tuple_row
from testcontainers.postgres import PostgresContainer

from psycopg import sql
from psyscale.async_core import ImproperInitilizationError
from psyscale.dev import *
from psyscale import PsyscaleAsync, PsyscaleConnectParams, PsyscaleDB
from psyscale.psql.enum import GenericTbls, Schema
from psyscale.psql.generic import list_schemas


# Test naming convention done to ensure test orders
def test_00_initlization_state(test_url):
    # Check to make sure there is nothing in the database already
    with (
        connect(test_url) as _conn,
        _conn.cursor(row_factory=tuple_row) as _cursor,
    ):
        _cursor.execute(list_schemas())
        rsp = {v[0] for v in _cursor.fetchall()}
        schemas_present = rsp.intersection(v for v in Schema)
        # None of the schemas we create should be present yet.
        assert len(schemas_present) == 0


@pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
def test_01_client_initilization_from_env(test_container: PostgresContainer):
    os.environ["PSYSCALE_HOST"] = test_container.get_container_host_ip()
    os.environ["PSYSCALE_PORT"] = str(test_container.get_exposed_port(test_container.port))
    os.environ["PSYSCALE_USER"] = test_container.username
    os.environ["PSYSCALE_PASSWORD"] = test_container.password
    os.environ["PSYSCALE_DB_NAME"] = test_container.dbname

    db = PsyscaleDB()
    rtn, status = db.execute(sql.SQL("SELECT 1").format())

    assert rtn[0][0] == 1
    assert status == "SELECT 1"

    # Ensure vars are cleared and error is raised
    env_vars = [
        "PSYSCALE_DB_NAME",
        "PSYSCALE_HOST",
        "PSYSCALE_PASSWORD",
        "PSYSCALE_PORT",
        "PSYSCALE_USER",
        "PSYSCALE_VOLUME_PATH",
    ]

    for var in env_vars:
        os.environ.pop(var)

    with pytest.raises(AttributeError):
        # No Environment variables are set, should error
        db = PsyscaleDB()


@pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
def test_02_client_initilization_from_params(test_container: PostgresContainer):
    conn_params = PsyscaleConnectParams(
        test_container.get_container_host_ip(),
        test_container.get_exposed_port(test_container.port),
        test_container.username,
        test_container.password,
        test_container.dbname,
    )

    db = PsyscaleDB(conn_params)
    rtn, status = db.execute(sql.SQL("SELECT 1").format())

    assert rtn[0][0] == 1
    assert status == "SELECT 1"


@pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
def test_03_client_initilization_from_url(test_container: PostgresContainer):
    # Test passing Param Obj
    conn_params = PsyscaleConnectParams.from_url(test_container.get_connection_url())
    db = PsyscaleDB(conn_params)
    rtn, status = db.execute(sql.SQL("SELECT 1").format())

    assert rtn[0][0] == 1
    assert status == "SELECT 1"

    # test passing url directly
    conn_params = test_container.get_connection_url()
    db = PsyscaleDB(conn_params)
    rtn, status = db.execute(sql.SQL("SELECT 1").format())

    assert rtn[0][0] == 1
    assert status == "SELECT 1"


@pytest.fixture(scope="module")
def db(test_container: PostgresContainer):
    conn_params = PsyscaleConnectParams.from_url(test_container.get_connection_url())
    yield PsyscaleDB(conn_params)


@pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
def test_04_schema_initilization(db: PsyscaleDB):
    rsp, status = db.execute(db[Op.SELECT, GenericTbls.SCHEMA]())
    schemas = {v[0] for v in rsp}
    assert schemas.issuperset(v for v in Schema)
    assert str(status).startswith("SELECT")


def test_05_security_schema_generation(db: PsyscaleDB):
    rsp, status = db.execute(db[Op.SELECT, GenericTbls.SCHEMA_TABLES](Schema.SECURITY))
    tables: set[str] = {v[0] for v in rsp}
    assert tables.issuperset({AssetTbls.SYMBOLS})
    assert str(status).startswith("SELECT")

    rsp, status = db.execute(db[Op.SELECT, GenericTbls.VIEW](Schema.SECURITY))
    tables: set[str] = {v[0] for v in rsp}
    assert tables.issuperset({AssetTbls._METADATA})
    assert str(status).startswith("SELECT")

    # Check _metadata sub-function exists
    rsp, status = db.execute(
        sql.SQL(
            """SELECT EXISTS (
                SELECT 1 FROM pg_proc
                WHERE proname = 'get_timeseries_date_range'
            );"""
        ).format()
    )

    assert rsp[0][0] is True
    assert str(status).startswith("SELECT")

    # Check trigram symbol search sub-function exists
    rsp, status = db.execute(
        sql.SQL(
            """SELECT EXISTS (
                SELECT 1 FROM pg_extension
                WHERE extname = 'pg_trgm'
            );"""
        ).format()
    )

    assert rsp[0][0] is True
    assert str(status).startswith("SELECT")


@pytest.mark.asyncio(loop_scope="session")
@pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
async def test_06_async_init(test_url):
    # with pytest.raises(ImproperInitilizationError):
    #     PsyscaleAsync(test_url)

    # No Error Raised
    _db = PsyscaleAsync(test_url)
    assert await _db._async_health_check()
    await _db.close()
